﻿
/**
 *	Accept signal values as a string and generates onValue events by timer interval. 
 */
function BC_SignalLevelsPorcessor(onValue)
{	
	var fLastSignal			= 0.0;
	var aSignal				= [];
	var nSignalIntervalID	= -1;
	
	/**
	 *	Clears interval timer of media signal.
	 */
	function clearSignalInterval()
	{
		if (nSignalIntervalID != -1)
		{
			clearInterval(nSignalIntervalID);
			nSignalIntervalID != -1;
		}
	}
	
	/**
	 *	 Generates signal event, called by timer.
	 */
	function updateMediaLevel()
	{
		if (aSignal.length == 0)
		{
			fLastSignal *= 0.5;
			if (fLastSignal < 0.02)
			{
				fLastSignal = 0;
				clearSignalInterval();
			}
			onValue(fLastSignal);
			return;
		}
		onValue(aSignal.shift());
	}
	
	/**
	 *	Sets levels of signal.
	 *	@param levels	String that contains signal values.
	 *	@param interval	Interval, milliseconds.
	 */
	this.setSignalLevels = function(values, interval)
	{	
		var count = values.length / 2;
		/* stop interval*/
		clearSignalInterval();
		
		/* fill signal values*/
		if (aSignal.length > count/2) aSignal = [];
		for (var i = 0; i < count; i++)
			aSignal.push(parseInt(values.substr(i*2, 2), 16)/255);
		if (aSignal.length)
			fLastSignal = aSignal[aSignal.length -1];
		
		/* start signal values;*/
		if (count)
			nSignalIntervalID = setInterval(updateMediaLevel, interval);
		else
			onValue(0);
	};
}

/**
 *	Executes handler function no more often than after a specified interval.
 */
function BC_IntervalExecution(interval, handler)
{
	interval = interval || 100;
	var nLastTime		= 0;
	var nTimeoutID		= -1;
	
	function executeHandler()
	{
		nLastTime = Date.now();
		handler();
		nTimeoutID = -1;
	}
	
	this.execute = function()
	{
		if (nTimeoutID == -1)
		{
			var time = Date.now();
			if (time - nLastTime >= interval)
				executeHandler();
			else
				nTimeoutID = setTimeout(executeHandler, interval - (time - nLastTime));
		}
	};
}

/**
 *	Broadcaster class.
 *
 *	@param  encoderCodec	Name of the encoder codec: 'opus'(default), 'mp3'.
 *	@param	encoderPath		Path to opus encoder (encoderWorker.min.js by default)
 *	@param	decoderPath		Path to opus decoder (decoderWorker.min.js by default)
 */
function Broadcaster(encoderCodec, encoderPath, decoderPath)
{
	/*
	 *	PRIVATE
	 */
	
	/* capture context*/
	var _broadcaster = this;
	
	/* extends LiveConnection*/
	var _connection = new LiveConnection(onWebscoketConnect, onWebsocketMessage, onWebsocketData, 5000);
	
    var AudioContext			= window.AudioContext || window.webkitAudioContext;
    var nSampleRate				= 44100;
    
    /**
     * Initialization of audio context.
     */
    function createAudioContext() 
    {
        if (!AudioContext) return null;
        var useSampleRate = false;
        if (navigator.mediaDevices && typeof navigator.mediaDevices.getSupportedConstraints == 'function') {
            var c = navigator.mediaDevices.getSupportedConstraints();
            useSampleRate = c.sampleRate || false;
        }
        return navigator.userAgent.toLowerCase().indexOf('firefox') > -1 && !useSampleRate ?
            (new AudioContext()) :
            (new AudioContext({latencyHint: "interactive", sampleRate: nSampleRate}));
    }

    var objAudioContext			= createAudioContext();
    var objAnalyzer				= null;
    
    var aMicEncoders            = [
        { name: 'opus', codec: 'opus', sampleRate: 24000, bitRate: 96000 },
        { name: 'opus64kbit', codec: 'opus', numChannels: 1, sampleRate: 24000, bitRate: 64000 },
        { name: 'raw22', codec: 'raw', sampleRate: 22050, bufferLength: 1470 },
        /*{ name: 'raw19', codec: 'raw', sampleRate: 19500, bufferLength: 1300 },*/
        { name: 'raw16', codec: 'raw', sampleRate: 16000, bufferLength: 1000 },
    ];
    var sEncoderCodec			= (findEncoder(encoderCodec) ? encoderCodec : false) || 'opus';
	var objEncoder				= null;
	var bLogin					= false;
    var bAudioCapture			= false;
    var bCaptureStarted			= false;
    var bAutostartUnallowed     = true;
	var nBufferLength			= 4096;
    var nInputChannels			= 2;
    var nInputBuffers			= 2;
	var nOutputChannels			= 2;
	var objCapture				= null;
    var objWebCapture			= null;
    var objPlayMedia            = null;
    var sAutoBCState            = '';
    var bIsPlaylistChanged      = false;
	/* stream playback*/
	var bStreamPlay				= false;/*loadValue('stream_playback', false);*/
	var bIsMuted				= loadValue('stream_mute', false);
	var fStreamVolume			= loadValue('stream_volume', 1.0);
	var fMicVolume				= loadValue('mic_volume', 1.0);
	var fMediaVolume			= loadValue('media_volume', 1.0);
	var bTrackNameVisible		= true;
	var objSendMediaVolume		= new BC_IntervalExecution(100, function(){applyMediaVolume(fMediaVolume);});
	var fAutoDJVolume			= 1.0;
	var objSendAutoDJVolume		= new BC_IntervalExecution(100, function(){applyAutoDJVolume(fAutoDJVolume);});
	var objMute					= {};
	var objMediaSignal			= new BC_SignalLevelsPorcessor(function(val){ if (_broadcaster.onMediaSignal) _broadcaster.onMediaSignal(val);});
	var objAutoDJSignal			= new BC_SignalLevelsPorcessor(function(val){ if (_broadcaster.onAutoDJSignal) _broadcaster.onAutoDJSignal(val);});
	var sCurMediaCallback		= false;
	var objPlayerControl		= null;
	var bHearMyself				= loadValue('hear_myself', true);
    var bCallAccept				= false;
    var bPlaylistCrossfade      = loadValue('playlist_crossfade', true);
    var nCrossfadeDuration      = 6000;
    var nFadeDuration           = 1000;
	var nLastPosition			= 0;
	var sLastPosCallback		= null;
	var sStopCallback			= null;
	var objPlayer				= new StreamPlayer(objAudioContext, decoderPath);
    var objPlayerGainNode		= objAudioContext ? objAudioContext.createGain() : null;
    if (objPlayerGainNode) objPlayerGainNode.connect(objAudioContext.destination);
    applyStreamVolume(fStreamVolume);
    initEncoder();

	
	/**
	 * Saves varialbe to the local storage 
	 */
	function saveValue(name, value)
	{
		localStorage.setItem(name, JSON.stringify({"val":value}));
	}
	
	/**
	 * Loads varialbe from the local storage 
	 */
	function loadValue(name, defValue)
	{
		var val = localStorage.getItem(name);
		if (val === null) return defValue;
		try{ val = JSON.parse(val);}
		catch(e) { return val; }
		return val.val;
    }
    
    function findEncoder(name)
    {
        for (var i = 0; i < aMicEncoders.length; i++)
        {
            if (aMicEncoders[i].name == name) return aMicEncoders[i];
        }
        return false;
    }

    function initEncoder()
    {
        var encoderName = loadValue('mic_encoder', false);

        if (!findEncoder(encoderName))
        {
            var mobile = false;
            (function(a){if(/(android|bb\d+|meego).+mobile|avantgo|bada\/|blackberry|blazer|compal|elaine|fennec|hiptop|iemobile|ip(hone|od)|iris|kindle|lge |maemo|midp|mmp|mobile.+firefox|netfront|opera m(ob|in)i|palm( os)?|phone|p(ixi|re)\/|plucker|pocket|psp|series(4|6)0|symbian|treo|up\.(browser|link)|vodafone|wap|windows ce|xda|xiino|android|ipad|playbook|silk/i.test(a)||/1207|6310|6590|3gso|4thp|50[1-6]i|770s|802s|a wa|abac|ac(er|oo|s\-)|ai(ko|rn)|al(av|ca|co)|amoi|an(ex|ny|yw)|aptu|ar(ch|go)|as(te|us)|attw|au(di|\-m|r |s )|avan|be(ck|ll|nq)|bi(lb|rd)|bl(ac|az)|br(e|v)w|bumb|bw\-(n|u)|c55\/|capi|ccwa|cdm\-|cell|chtm|cldc|cmd\-|co(mp|nd)|craw|da(it|ll|ng)|dbte|dc\-s|devi|dica|dmob|do(c|p)o|ds(12|\-d)|el(49|ai)|em(l2|ul)|er(ic|k0)|esl8|ez([4-7]0|os|wa|ze)|fetc|fly(\-|_)|g1 u|g560|gene|gf\-5|g\-mo|go(\.w|od)|gr(ad|un)|haie|hcit|hd\-(m|p|t)|hei\-|hi(pt|ta)|hp( i|ip)|hs\-c|ht(c(\-| |_|a|g|p|s|t)|tp)|hu(aw|tc)|i\-(20|go|ma)|i230|iac( |\-|\/)|ibro|idea|ig01|ikom|im1k|inno|ipaq|iris|ja(t|v)a|jbro|jemu|jigs|kddi|keji|kgt( |\/)|klon|kpt |kwc\-|kyo(c|k)|le(no|xi)|lg( g|\/(k|l|u)|50|54|\-[a-w])|libw|lynx|m1\-w|m3ga|m50\/|ma(te|ui|xo)|mc(01|21|ca)|m\-cr|me(rc|ri)|mi(o8|oa|ts)|mmef|mo(01|02|bi|de|do|t(\-| |o|v)|zz)|mt(50|p1|v )|mwbp|mywa|n10[0-2]|n20[2-3]|n30(0|2)|n50(0|2|5)|n7(0(0|1)|10)|ne((c|m)\-|on|tf|wf|wg|wt)|nok(6|i)|nzph|o2im|op(ti|wv)|oran|owg1|p800|pan(a|d|t)|pdxg|pg(13|\-([1-8]|c))|phil|pire|pl(ay|uc)|pn\-2|po(ck|rt|se)|prox|psio|pt\-g|qa\-a|qc(07|12|21|32|60|\-[2-7]|i\-)|qtek|r380|r600|raks|rim9|ro(ve|zo)|s55\/|sa(ge|ma|mm|ms|ny|va)|sc(01|h\-|oo|p\-)|sdk\/|se(c(\-|0|1)|47|mc|nd|ri)|sgh\-|shar|sie(\-|m)|sk\-0|sl(45|id)|sm(al|ar|b3|it|t5)|so(ft|ny)|sp(01|h\-|v\-|v )|sy(01|mb)|t2(18|50)|t6(00|10|18)|ta(gt|lk)|tcl\-|tdg\-|tel(i|m)|tim\-|t\-mo|to(pl|sh)|ts(70|m\-|m3|m5)|tx\-9|up(\.b|g1|si)|utst|v400|v750|veri|vi(rg|te)|vk(40|5[0-3]|\-v)|vm40|voda|vulc|vx(52|53|60|61|70|80|81|83|85|98)|w3c(\-| )|webc|whit|wi(g |nc|nw)|wmlb|wonu|x700|yas\-|your|zeto|zte\-/i.test(a.substr(0,4)))
                mobile = true;})(navigator.userAgent||navigator.vendor||window.opera);
            var nCores = typeof navigator.hardwareConcurrency == 'number' ? navigator.hardwareConcurrency : 1;
            var slowConnection = false;
            var conn = navigator.connection || navigator.webkitConnection || navigator.mozConnection || navigator.msConnection;
            if (conn) 
            {
                var t = conn.effectiveType;
                var dl = conn.downlink;
                if ((dl && dl < 0.8) || /* bandwidth < 1 Mbit/s*/
                    (t && (t == '3g' || t == '2g' || t == 'slow-2g'))) /* 3g or worse*/
                    slowConnection = true;
            }
            if (slowConnection) console.log('Connection is slow!');

            if (mobile)
            {
                sEncoderCodec = nCores > 4 ? 'opus64kbit' : 'raw16';
            }
            else /* desktop*/
            {
                if (nCores > 2) 
                {
                    sEncoderCodec = slowConnection ? 'opus64kbit' : 'opus';
                }
                else
                {
                    sEncoderCodec = slowConnection ? 'raw16' : 'raw22';
                }
            }
        }
        else
        {
            sEncoderCodec = encoderName;
        }
        
        setTimeout(function() { _broadcaster.onSelectMicEncoder(sEncoderCodec); }, 100);
    }

	
	/**
	 * Sets stream volume.
	 */
	function applyStreamVolume(value) { if (objPlayerGainNode) objPlayerGainNode.gain.value = bIsMuted ? 0 : value; }
	
	/**
	 *	Handler of encoded data. 
	 */
	function onDataEncoded(data)
	{
		if (data.length && objEncoder) _connection.send(data);
	}
	
	/**
	 *	Returns control object of audio capture. 
	 */
	function getWebCapture()
	{
		if (objWebCapture === null)
			objWebCapture = new WebAudioCapture();
		return objWebCapture;
	}
	
	/**
	 *	Processes audio stream.
	 */
	function onAudioData(buffers)
	{
		if (!bCaptureStarted)
		{
            if (!bAudioCapture || !_connection.isConnected()) return;
            nInputBuffers = buffers.length;
            if (!startEncoder(nInputBuffers)) return;
            bCaptureStarted = true;
			if (_broadcaster.onStartCapture) _broadcaster.onStartCapture();
		}
		
		if (_broadcaster.onMicSignal)
		{
			/* calculate signal level*/
			var aData = buffers[0];
			var sum = 0, max = 0;
			for (var i = 0; i < nBufferLength - 3; i += 4)
			{
				sum = (sum + Math.abs(aData[i]) + Math.abs(aData[i + 2])) / 3;
				if (sum > max) max = sum; 
			}
			_broadcaster.onMicSignal(max);
		}
		
		if (objEncoder) objEncoder.encode(buffers);
    }
    
    /**
     * Init encoder
     * @param {*} inputChannels     Number of input buffers
     */
    function startEncoder(inputChannels)
    {
		var encoder = findEncoder(sEncoderCodec);
		if (!encoder) encoder = aMicEncoders[0];		
        switch (encoder.codec)
        {
            case 'mp3':
                objEncoder = new LameEncoderHelper(onDataEncoded);
                objEncoder.init({numberOfChannels: inputChannels, sampleRate:nSampleRate, bitRate:128});
                _connection.send(JSON.stringify({cmd:'start_mic', format:'mp3'}));
            break;
            
            case 'opus':
                objEncoder = new OpusEncoderHelper(onDataEncoded, encoderPath);
                objEncoder.init({
                    numberOfChannels: encoder.numChannels ? encoder.numChannels : inputChannels,
                    originalSampleRate: nSampleRate,
                    sampleRate: encoder.sampleRate,
                    bitRate: encoder.bitRate,
                    bufferLength: nBufferLength
                });
                _connection.send(JSON.stringify({cmd:'start_mic', format:'ogg'}));
            break;
            
            case 'raw':
                objEncoder = new RawEncoderHelper(onDataEncoded);
                objEncoder.init({
                    numberOfChannels: inputChannels,
                    originalSampleRate: nSampleRate,
                    sampleFormat: 'S16', /* 'U8'*/
                    sampleRate: encoder.sampleRate,
                    bufferLength: encoder.bufferLength
                });
                _connection.send(JSON.stringify({cmd:'start_mic', format:objEncoder.getFormat()}));
            break;

            default:
                _broadcaster.stopAudioCapture();
                return false;
        }
        return true;
    }
	
	/**
	 *	Handler of connection/disconnection.
	 *	@param connected	Boolean flag.
	 */
	function onWebscoketConnect(connected)
	{
		if (connected)
		{
			_connection.send(JSON.stringify({cmd:'init_connection', type:'broadcast'}));
		}
		else
		{
            bLogin = false;
            objEncoder = null;
            /*_broadcaster.stopAudioCapture(true);*/
            if (_broadcaster.onRecordStatus) _broadcaster.onRecordStatus('disconnect', '');
            if (_broadcaster.onDisconnect) _broadcaster.onDisconnect();
            sAutoBCState = '';
            _broadcaster.setPlaylistChanged(bIsPlaylistChanged);
		}
    }
    
	/**
	 *	Processes server commands.
	 *	@param msg	A message object, received from the server.
	 */
	function onWebsocketMessage(msg)
	{
		switch (msg.cmd)
		{
			case 'set_salt':
				if (_broadcaster.onLoginRequest) _broadcaster.onLoginRequest(msg.salt);
			break;
			
			case 'login':
				bLogin = (msg.status == 'OK');
				if (_broadcaster.onLogin) _broadcaster.onLogin(msg.status);
				if (bCallAccept) _broadcaster.setCallsAccept(true); 
				if (bLogin)
				{
                    _connection.send(JSON.stringify({cmd:'set_playlist_crossfade', crossfade:(bPlaylistCrossfade ? nCrossfadeDuration : 0), fade:nFadeDuration}));
                    if ((bStreamPlay || objPlayerControl && objPlayerControl.isAutoStart())) _broadcaster.playStream(bHearMyself);
                    if (bCaptureStarted) startEncoder(nInputBuffers);
                    /* restore Auto DJ volume*/
                    if (fAutoDJVolume != 1.0 || objMute.autodj)
                    {
                        applyAutoDJVolume(objMute.autodj ? 0 : fAutoDJVolume);
                    }
                    if (fMediaVolume != 1.0 || objMute.media)
                    {
                        applyMediaVolume(objMute.media ? 0 : fMediaVolume);
                    }
                    if (objPlayMedia)
                    {
                        objPlayMedia.position = nLastPosition;
                        _connection.send(JSON.stringify(objPlayMedia));
                    }
                }
                else
                {
                    objPlayMedia = null;
                    _broadcaster.stopAudioCapture();
                    fAutoDJVolume = 1.0;
                    objMute.autodj = false;
                    if (_broadcaster.onAutoDJVolumeChange) _broadcaster.onAutoDJVolumeChange(msg.volume);
                    _broadcaster.onMute('autodj', false);
                }
            break;
            
            case 'start_stream':
                if (msg.status != 'OK')
                {
                    console.error('Start stream of mic failed', msg);
                    _broadcaster.stopAudioCapture();
                }
            break;
			
			case 'play_media':
				if (_broadcaster.onPlayMedia) {
                    _broadcaster.onPlayMedia(msg.status, msg.callback, msg.duration, msg.error);
                    if (objPlayMedia && objPlayMedia.callback == msg.callback && 
                        msg.status != 'OK' && msg.error != 'already_started')
                    {   
                        objPlayMedia = null;
                    }
                }
			break;
			
			case 'stop_media':
				if (sLastPosCallback != sStopCallback && sLastPosCallback == msg.callback) nLastPosition = 0;
				if (_broadcaster.onStopMedia) _broadcaster.onStopMedia(msg.status, msg.callback, msg.error);
            break;
            
            case 'start_crossfade':
                if (_broadcaster.onStopMedia) _broadcaster.onStopMedia('CROSSFADE', msg.callback, null);
            break;
			
			case 'update_pos':
				nLastPosition = msg.pos;
				sLastPosCallback = msg.callback; 
				if (_broadcaster.onUpdatePos) _broadcaster.onUpdatePos(msg.callback, msg.pos);
				objMediaSignal.setSignalLevels(msg.signal, msg.interval);
			break;
			
			case 'seek':
				if (_broadcaster.onSeek) _broadcaster.onSeek(msg.callback);
            break;
            
            case 'update_autodj_indexes':
                if (_broadcaster.onUpdateAutoDJIndexes) _broadcaster.onUpdateAutoDJIndexes(msg.indexes);
            break;
			
			case 'set_autobc_state':
                sAutoBCState = msg.state;
                if (_broadcaster.onAutoBCState) _broadcaster.onAutoBCState(msg);
                _broadcaster.setPlaylistChanged(bIsPlaylistChanged);
            case 'set_autobc_volume': /* || 'set_autobc_state'*/
                if (bLogin || fAutoDJVolume == 1.0 && !objMute.autodj) /* if no changes of AutoDJ volume*/
                {
				    if (msg.hasOwnProperty('volume') && (!objMute.autodj || msg.volume != 0)) 
                    {
                        fAutoDJVolume = msg.volume;
                        if (_broadcaster.onAutoDJVolumeChange) _broadcaster.onAutoDJVolumeChange(msg.volume);
                        if (objMute.autodj) { objMute.autodj = false; _broadcaster.onMute('autodj', false); }
                    }
                }
            break;
            
            case 'sync_autobc_playlist':
                if (_broadcaster.onSyncAutoBCPlaylist) _broadcaster.onSyncAutoBCPlaylist(msg);
            break;
			
			case 'set_autobc_signal':
				objAutoDJSignal.setSignalLevels(msg.signal, msg.interval);
			break;
			
			case 'reset_stream':
				if (bStreamPlay)
				{
					objPlayer.stop();
					objPlayer.start(objAnalyzer === null ? objPlayerGainNode : objAnalyzer);
				}
			break;
			
			case 'set_desc':
				if (msg.hasOwnProperty("desc"))
				{
					if (objPlayerControl) objPlayerControl.onDescriptionChanged(msg.desc);
					if (_broadcaster.onDescriptionChanged) _broadcaster.onDescriptionChanged(msg.desc);
				}
			break;
			
			case 'set_rec_status':
				if (_broadcaster.onRecordStatus) _broadcaster.onRecordStatus(msg.status, msg.desc);
			break;
			
			case 'set_rec_pos':
				if (_broadcaster.onRecordDuration) _broadcaster.onRecordDuration(msg.pos, msg.max);
			break;
			
			case 'record_saved':
				if (_broadcaster.onSaveRecord) _broadcaster.onSaveRecord(msg.file_name);
			break;
			
			case 'counter':
				if (_broadcaster.onListenersCounter) _broadcaster.onListenersCounter(msg.val, msg.max);
			break;
			
			case 'bc_calls_accept':
				if (_broadcaster.onAcceptCalls) _broadcaster.onAcceptCalls(msg.val);
				if (msg.calls && _broadcaster.onAddCall)
				{
					for (var i = 0; i < msg.calls.length; i++)
						_broadcaster.onAddCall(msg.calls[i]);
				}
			break;
			
			case 'add_call':
				if (_broadcaster.onAddCall) _broadcaster.onAddCall(msg.call);
			break;
			
			case 'remove_call':
				if (_broadcaster.onRemoveCall) _broadcaster.onRemoveCall(msg.id);
			break;
			
			case 'update_call':
				if (_broadcaster.onUpdateCall) _broadcaster.onUpdateCall(msg.id, msg.status);
            break;
		}
	}
	
	/**
	 *	Processes binary data.
	 *	@param data		Bitary data, received from the server.
	 */
	function onWebsocketData(data)
	{
		/* binary data (stream)*/
		if (bStreamPlay) objPlayer.playData(data);
	}
	
	/*
	 *	PUBLIC
	 */
	 
	/* events*/
	this.onLoginRequest			= null;
	this.onLogin				= null;
	this.onDisconnect			= null;
	this.onStartCapture			= null;
	this.onStopCapture			= null;
    this.onCaptureError			= null;
    this.onSelectMicEncoder     = function() {};
	this.onPlayMedia			= null;
	this.onStopMedia			= null;
	this.onMediaVolumeChange	= null;
	this.onUpdatePos			= null;
    this.onSeek					= null;
	this.onMicSignal			= null;	/* Called during the broadcast from mic to display sound waves*/
	this.onMediaSignal			= null; /* Called during the signal of broadcasted media volume*/
	this.onAutoDJSignal			= null;
	this.onAutoBCState			= null;
    this.onAutoDJVolumeChange	= null;
    this.onUpdateAutoDJIndexes  = null;
    this.onSetAutoDJChanged     = null;
    this.onSyncAutoBCPlaylist   = null;
	this.onMute					= function(bMuted) {};
	/* stream playback events*/
	this.onPlay					= function(bPlay) {};
	this.onStartVisualizer		= null;
	this.onStopVisualizer		= null;
	this.onBufferEmpty			= function() {};
	this.onBufferFull			= function() {};
	this.onListenersCounter		= function() {};
	this.onBufferOverflow		= function() {};
	this.onDescriptionChanged	= function() {};
	this.onStreamMute			= function(bMuted) {};
	this.onVolumeChanged		= function(fVolume) {};
	/* record events*/
	this.onRecordStatus			= null;
	this.onRecordDuration		= null;
	this.onSaveRecord			= null;
	/* calls*/
	this.onAcceptCalls			= null;
	this.onAddCall				= null;
	this.onRemoveCall			= null;
	this.onUpdateCall			= null;
	objPlayer.onBufferEmpty = function(){
		if (objPlayerControl) objPlayerControl.onBuffering(true);
		if (_broadcaster.onBufferEmpty) _broadcaster.onBufferEmpty(); 
	};
	objPlayer.onBufferFull = function(){
		if (objPlayerControl) objPlayerControl.onBuffering(false);
		if (_broadcaster.onBufferFull) _broadcaster.onBufferFull();
	};
	objPlayer.onBufferOverflow = function() { if (_broadcaster.onBufferOverflow) _broadcaster.onBufferOverflow(); };
	
	/**
	 * Initializes broadcaster. 
	 */
	this.initialize = function()
	{
		_broadcaster.onPlay(bStreamPlay);
		_broadcaster.onStreamMute(bIsMuted);
		_broadcaster.onVolumeChanged(fStreamVolume);
	};
	
	/**
	 *	Initializes connection.
	 */
	this.connect = function(URL)
	{
		_connection.connect(URL);
	};
	
	/**
	 *	Resets a connection.
	 */
	this.reconnect = function()
	{
		_connection.disconnect(true);
	};
	
	/**
	 *	Closes a connection.
	 */
	this.disconnect = function()
	{
		if (objPlayerControl)
		{
			objPlayerControl.setBroadcaster(null);
			objPlayerControl.connect(_connection.getURL());
			objPlayerControl = null;
		}
		_connection.disconnect();
	};
	
	/**
	 *	Sends login data.
	 */
	this.login = function(login, hash)
	{
		if (!_connection.isConnected() || bLogin)
			return false;
		_connection.send(JSON.stringify({cmd:'login', login:login, hash:hash}));
		return true;
    };
    
    /**
	 *	Is broadcaster logged in
	 */
	this.isLoggedIn = function()
	{
        return _connection.isConnected() && bLogin;
    };
	
	/**
	 *	Starts capture of an audio stream.
	 */
	this.startAudioCapture = function(captureObject)
	{
		if (bAudioCapture) return;
		if (!_connection.isConnected() || !bLogin)
		{
			if (_broadcaster.onCaptureError) _broadcaster.onCaptureError('not_connected');
			return;
        }
        if (bAutostartUnallowed)
        {
            bAutostartUnallowed = false;
            try 
            {
                objAudioContext.resume();
            }
            catch (err)
            {
                console.log(err);
            }
        }
		bAudioCapture = true;
		objCapture = captureObject || getWebCapture();
		if (!objCapture.startAudioCapture({
				audioContext: objAudioContext,
				bufferLength: nBufferLength,
				numInputChannels: nInputChannels,
				numOutputChannels: nOutputChannels,
				volume: objMute.mic ? 0 : fMicVolume,
				/* events*/
				onAudioData: onAudioData,
				onStartCapture: function(sampleRate){
						if (!bAudioCapture || !_connection.isConnected()) return false;
						nSampleRate = sampleRate;
						console.log("Start audio capture " + nSampleRate);
						return true;
					},
				onStopCapture: function(){
						if (bCaptureStarted){
							_connection.send(JSON.stringify({cmd:'stop_mic'}));
							bCaptureStarted = false;
							if (_broadcaster.onStopCapture) _broadcaster.onStopCapture();
						}
					},
				onErrorCapture: function(error, event){
                        bAudioCapture = false;
						if (_broadcaster.onCaptureError) _broadcaster.onCaptureError(error, event ? event.message : false);
					}
				}))
			{if (_broadcaster.onCaptureError) _broadcaster.onCaptureError('not_supported');}
	};
	
	/**
	 *	Stops capture of an audio stream.
	 */
	this.stopAudioCapture = function()
	{
        if (!bAudioCapture) return;
        bAudioCapture = false;
		objCapture.stopAudioCapture();
		if (_broadcaster.onMicSignal) _broadcaster.onMicSignal(0);
    };

    /**
     * Returns available encoders
     */
    this.getMicEncoders = function()
    {
        var names = [];
        aMicEncoders.forEach(function(codec) { names.push(codec.name); });
        return names;
    };

    /**
     * Returns current encoder
     */
    this.getMicEncoder = function() { return sEncoderCodec; };

    /**
     * Set current mic encoder by name
     */
    this.setMicEncoder = function(name)
    {
        if (!findEncoder(name)) return false;
        sEncoderCodec = name;
        console.log('save mic_encoder', sEncoderCodec);
        saveValue('mic_encoder', sEncoderCodec);
        this.onSelectMicEncoder(name);
        return true;
    };
	
	/**
	 *	Starts playback of a media file.
	 */
	this.playMedia = function(files, callback, title, author, position)
	{
        if (!_connection.isConnected() || !bLogin) return false;
		sCurMediaCallback = callback;
		sStopCallback = null;
        _broadcaster.setMediaVolume(fMediaVolume);
        
        if (!files.length)
            return false;
        
        objPlayMedia = {cmd:'play_media', file_name:files[0].file, url:files[0].url, callback:callback, title:title, author:author,
                        position:(position? parseInt(position) : 0), visible:bTrackNameVisible};

        if (files.length > 1)
        {
            objPlayMedia.backup_file_name = files[1].file;
            objPlayMedia.backup_url = files[1].url;
        }
        
		_connection.send(JSON.stringify(objPlayMedia));
		return true;
	};
	
	/**
	 *	Stops a media file playback.
	 */
	this.stopMedia = function(callback)
	{
        sStopCallback = callback;
        objPlayMedia = null;
        if (!_connection.isConnected() || !bLogin) 
        {
            if (_broadcaster.onStopMedia) _broadcaster.onStopMedia('OK', callback);
        }
        else
        {
            _connection.send(JSON.stringify({cmd:'stop_media', callback:callback}));
        }
		return true;
	};
	
	/**
	 *	Seek a media position.
	 *	@param callback	
	 */
	this.seekMedia = function(callback, position)
	{
        /*console.log('>>>', sCurMediaCallback, callback, position);*/
        if (!_connection.isConnected() || !bLogin || !callback)
        {
            if (sCurMediaCallback)
            {
                nLastPosition = position;
                if (_broadcaster.onUpdatePos) _broadcaster.onUpdatePos(callback, position);
            }
        }
        else
        {
            _connection.send(JSON.stringify({cmd:'seek_media', callback:callback, pos:position}));
        }
		return true;
    };
    
    /**
     * Enables/disables playlist crossfade
     */
    this.setPlaylistCrossfade = function(value)
    {
        bPlaylistCrossfade = value;
        saveValue('playlist_crossfade', bPlaylistCrossfade);
        _connection.send(JSON.stringify({cmd:'set_playlist_crossfade', crossfade:(bPlaylistCrossfade ? nCrossfadeDuration : 0)}));
    };

    this.getPlaylistCrossfade = function() { return bPlaylistCrossfade; };
	
	/**
	 *	Starts playback of the station stream.
	 */
	this.playStream = function()
	{
		if (_connection.send(JSON.stringify({cmd:'play_stream', hear_myself:bHearMyself})))
		{
			if (_broadcaster.onStartVisualizer)
			{
				objAnalyzer = objAudioContext.createAnalyser();
				objAnalyzer.connect(objPlayerGainNode);
				objPlayer.start(objAnalyzer);
				_broadcaster.onStartVisualizer(objAnalyzer);
			}
			else
			{
				objPlayer.start(objPlayerGainNode);
			}
			if (objPlayerControl) objPlayerControl.onPlay(true);
			saveValue('stream_playback', true);
			bStreamPlay = true;
			_broadcaster.onPlay(true);
			return true;
		}
		return false;
	};
	
	/**
	 *	Stops playback of the station stream.
	 */
	this.stopStream = function()
	{
		if (objAnalyzer !== null)
		{
			if (_broadcaster.onStopVisualizer) _broadcaster.onStopVisualizer();
			objAnalyzer.disconnect();
			objAnalyzer = null;
		}
		objPlayer.stop();
		if (objPlayerControl) objPlayerControl.onPlay(false);
		saveValue('stream_playback', false);
		bStreamPlay = false;
		_broadcaster.onPlay(false);
		return _connection.send(JSON.stringify({cmd:'stop_stream'}));
	};
	
	/**
	 * Set mute state
	 * @param bMute
	 */
	this.setStreamMute = function(bMute)
	{
		saveValue('stream_mute', bMute);
		bIsMuted = bMute;
		_broadcaster.onStreamMute(bIsMuted);
		applyStreamVolume(fStreamVolume);
	};
	
	/**
	 *	Sets hear myself option. 
	 */
	this.setHearMyself = function(hearMyself)
	{
		bHearMyself = hearMyself;
		saveValue('hear_myself', bHearMyself);
		return _connection.send(JSON.stringify({cmd:'set_hear_myself', hear_myself:bHearMyself}));
	};
	
	this.getHearMyself = function() { return bHearMyself; };
	
	/**
	 *	Returns true if "hear myself" option is enabled.  
	 */
	this.isHearMyself = function() { return bHearMyself; };
	
	/**
	 *	Sets player stream volume. 
	 *	@param value	Volume value [0..1].
	 */
	this.setStreamVolume = function(value)
	{
		if (bIsMuted)
		{
			saveValue('stream_mute', false);
			bIsMuted = false;
			_broadcaster.onStreamMute(false);
		}
		saveValue('stream_volume', value);
		fStreamVolume = value;
		_broadcaster.onVolumeChanged(fStreamVolume);
		applyStreamVolume(value);
	};
	
	/**
	 *	Sets microphone volume.
	 *	@param value	Volume value [0..1]. 
	 */
	this.setMicVolume = function(value)
	{
		fMicVolume = value;
		applyMicVolume(value);
		if (objMute.mic) _broadcaster.setMute('mic', false);
	};
	
	function applyMicVolume(value)
	{
		if (objWebCapture) objWebCapture.setMicVolume(value);	
	}
	
	/**
	 *	Sets media volume.
	 *	@param value	Volume value [0..1]. 
	 */
	this.setMediaVolume = function(value)
	{
		fMediaVolume = value;
		objSendMediaVolume.execute();
		if (_broadcaster.onMediaVolumeChange) _broadcaster.onMediaVolumeChange(sCurMediaCallback, fMediaVolume);
		if (objMute.media) _broadcaster.setMute('media', false);
	};
	
	function applyMediaVolume(value)
	{
		_connection.send(JSON.stringify({cmd:'set_media_volume', value:value}));
	}
	
	/** Media volume accessor. */
	this.getMediaVolume = function() { return fMediaVolume; };
	
	/**
	 *	Stops auto broadcaster.
	 */
	this.stopAutoBC = function()
	{
		if (!_connection.isConnected() || !bLogin) return false;
		_connection.send(JSON.stringify({cmd:'stop_autobc'}));
		return true;
	};
	
	/**
	 *	Sets autobroadcaster parameters.
	 *	@param params Object {shuffle:true|false, crossfade:"slow"|"fast", suspension:true|false}
	 */
	this.sendAutoBCParams = function(params, start)
	{
		if (!_connection.isConnected() || !bLogin) return false;
		var msg = {cmd: (start ? 'start_autobc': 'set_autobc_params')};
		if (typeof(params.shuffle) != 'undefined') msg.shuffle = params.shuffle ? true : false;
		if (typeof(params.crossfade) != 'undefined') msg.crossfade = params.crossfade;
        if (typeof(params.suspension) != 'undefined') msg.suspension = params.suspension ? true : false;
        if (typeof(params.index) != 'undefined' && params.index !== false) msg.index = params.index;
		_connection.send(JSON.stringify(msg));
		return true;
	};
	
	/*Starts auto broadcaster.
	*@param shuffle		boolean
	*@param crossfade	string	'slow' of 'fast'
	*/
	this.startAutoBC = function(shuffle, crossfade)
	{
		if (!_connection.isConnected() || !bLogin) return false;
		_connection.send(JSON.stringify({cmd:'start_autobc', shuffle:(shuffle ? true : false), crossfade:crossfade}));
		return true;
	};
    
    /**
     *  Apply changes of auto DJ playlist
     */
    this.applyAutoBCPlaylist = function(index, playlist)
    {
        if (!_connection.isConnected() || !bLogin) return false;
        var msg = {cmd:'apply_autobc_playlist'};
        if (index !== false) msg.index = index;
        if (playlist !== false) msg.playlist = playlist;
		_connection.send(JSON.stringify(msg));
		return true;
    };

    /**
     *  Start playback from index
     */
    this.startAutoBCIndex = function(index)
    {
        if (!_connection.isConnected() || !bLogin) return false;
		_connection.send(JSON.stringify({cmd:'start_autobc_index', 'index':index}));
		return true;
    };
	
	/**
	 *	Sets auto DJ sound volume.
	 */
	this.setAutoDJVolume = function(volume)
	{
		fAutoDJVolume = volume;
		objSendAutoDJVolume.execute();
		if (objMute.autodj) _broadcaster.setMute('autodj', false);
	};
	
	function applyAutoDJVolume(value)
	{
		_connection.send(JSON.stringify({cmd:'set_autobc_volume', value:value}));
	}
	
	/**
	 *	 Sets player to control stream playback.
	 */
	this.setPlayer = function(player)
	{
		objPlayerControl = player;
		if (objPlayerControl)
		{
			objPlayerControl.disconnect();
			objPlayerControl.setBroadcaster(_broadcaster);
		}
	};
	
	/**
	 *	Starts recording.
	 */
	this.startRecord = function(title, path)
	{
		if (!_connection.isConnected() || !bLogin) return false;
		if (typeof(path) == 'undefined') path = ''; 
		_connection.send(JSON.stringify({cmd:'start_record', title:title, path: path}));
		return true;
	};
	
	/**
	 *	Stops recording.
	 */
	this.stopRecord = function(title)
	{
		if (!_connection.isConnected() || !bLogin) return false;
		_connection.send(JSON.stringify({cmd:'stop_record', title:title}));
		return true;
	};
	
	/**
	 *	Sets air description.
	 */
	this.setAirDesc = function(desc)
	{
		if (!_connection.isConnected() || !bLogin) return false;
		_connection.send(JSON.stringify({cmd:'set_air_desc', desc:desc}));
		return true;
	};
	
	/**
	 *	Sets calls accept.
	 */
	this.setCallsAccept = function(bAccept)
	{
		if (!_connection.isConnected() || !bLogin) return false;
		bCallAccept = bAccept ? true : false;
		_connection.send(JSON.stringify({cmd:'set_calls_accept', accept:bCallAccept}));
		return true;
	};
	
	/**
	 *	Sets a new status of the call.
	 */
	this.setCallStatus = function(id, status)
	{
		if (!_connection.isConnected() || !bLogin) return false;
		_connection.send(JSON.stringify({cmd:'set_call_status', id:id, status:status}));
		return true;
	};
	
	/**
	 *	Stop the application.
	 */
	this.shutdownApp = function()
	{
		if (!_connection.isConnected() || !bLogin) return false;
		_connection.send(JSON.stringify({cmd:'shutdown'}));
		return true;
	};
	
	this.getFlag = function(name)
	{
		var objFlags = loadValue('flags', null);
		if (objFlags == null || !objFlags.hasOwnProperty(name))
		{
			if (name == 'playlist_goto_next') return true;
			return false;
		}
		return objFlags[name];
	};
	
	this.setFlag = function(name, value)
	{
		var objFlags = loadValue('flags', null);
		if (objFlags == null) objFlags = {};
		objFlags[name] = value ? true : false;
		saveValue('flags', objFlags);
	};
	
	this.getMediaPosition = function(callback)
	{
		return callback == sLastPosCallback ? nLastPosition : 0;
	};
	
	/**
	 * Sets muted mode for specified source type.
	 * @param type {string}		Source type to mute/unmute: 'mic', 'autodj' or 'media'.
	 * @param muted	{boolen}	Muted value.
	 */
	this.setMute = function(type, muted)
	{
		switch(type)
		{
			case 'mic':		applyMicVolume(muted ? 0 : fMicVolume); break;
			case 'autodj':	applyAutoDJVolume(muted ? 0 : fAutoDJVolume); break;
			case 'media':	applyMediaVolume(muted ? 0 : fMediaVolume); break;
		}
		objMute[type] = muted ? true : false;
		_broadcaster.onMute(type, muted);
	};
	
	/** Returns visibility of track name. */
	this.getTrackNameVisibility = function(){ return bTrackNameVisible; };
	
	/**
	 *	Sets visibility of track name.
	 */
	this.setTrackNameVisibility = function(visible)
	{
		if (!_connection.isConnected() || !bLogin) return false;
		bTrackNameVisible = !!visible;
		_connection.send(JSON.stringify({cmd:'set_track_name_visibility', value:bTrackNameVisible}));
		return true;
    };
    
    /**
     *  Set playlist changed status
     */
    this.setPlaylistChanged = function(isChanged)
    {
        bIsPlaylistChanged = isChanged;
        if (_broadcaster.onSetAutoDJChanged)
        {
            _broadcaster.onSetAutoDJChanged(bIsPlaylistChanged, sAutoBCState);
        }
    };

    /** Auto DJ state getter */
    this.getAutoBCState = function() { return sAutoBCState; };

    /**
     * Synchronize AutoBC playlist
     */
    this.syncAutoBCPlaylist = function(action, startedId, currentId)
    {
        if (!_connection.isConnected() || !bLogin) return false;
		_connection.send(JSON.stringify({cmd:'sync_autobc_playlist', 'action':action, 'started_id':startedId, 'current_id':currentId}));
		return true;
    };
}
