﻿/**
 *	Chat class.
 */
function Chat()
{
	/* capture context*/
	var _chat					= this;
	
	/* extends LiveConnection*/
	var _connection = new LiveConnection(onWebscoketConnect, onWebsocketMessage, null, 5000);
	
	/*
	 *	PRIVATE
	 */
	var sAppID				= 0;
	var aHiddenMessages		= [];
	var aMutedUsers			= (getItem('vchat_muted_users') || '').split(',');
	var sGUID				= getItem('vchat_guid');
	var sNick				= getItem('vchat_nick');
	var bBannedVisible		= getItem('vchat_show_banned') == 'false' ? false : true;
	var sSalt				= '';
    var aUsers				= []; /* sorted array of users*/
    var aBanned             = {};
	var objThisUser			= null;
	var aPrivateCounters	= [];
	var aPrivateMessages	= [];
	var aPollResults		= getPollResults();
	var bShowPollResults	= true;
	var bModeratorVisible	= false;
	var nMaxHistorySize		= 100;
	var nTotalUnreadCounter = 0;
    var objPoll				= null;
    
    function getnerateLocalId()
    {
        var msgGUID = ((new Date()).getTime() % 0xFFFFFFFF).toString(16) +'-'+ (Math.floor(Math.random()*16384)).toString(16);
        var msgInd = getItem('vchat_msg_ind') || 0x10;
        msgInd++;
        if (msgInd >= 0xFF) msgInd = 0x10;
        setItem('vchat_msg_ind', msgInd);
        return msgInd.toString(16) +'-'+ msgGUID +'-'+ sSalt;
    }

    function isBannedShow() { return bBannedVisible && _chat.isBanAllowed(); }

    function showBanned(show)
    {
        for (var id in aBanned)
        {
            if (aBanned.hasOwnProperty(id)) {
                if (show) addUser(aBanned[id]); else removeUser(aBanned[id]);
            }
        }
    }
	
	/**
	 * Checks if a message is hidden. 
	 */
	function isMessageHidden(msg)
	{
		return ((typeof(msg.id) != 'undefined' && aHiddenMessages.indexOf(parseInt(msg.id, 10)) != -1) || aMutedUsers.indexOf(msg.uid) != -1);
	}
	
	/**
	 * Checks if a user is muted. 
	 */
	function isUserMuted(userID)
	{
		return aMutedUsers.indexOf(userID) != -1;	
	}
	
	/**
	 *	Processes a message received from the server. 
	 */
	function addMessage(msg)
	{
		if (msg.uid) msg.uid = msg.uid;
		if (_chat.onAddMessage)
		{
			if (isMessageHidden(msg)) return;
			
			if (msg.target_uid)
			{
				msg.private = true;
				msg.target_id = msg.target_uid;
			}
			
			if (!_chat.onAddMessage(msg))
			{
				if (msg.private)
				{
					if (objThisUser && msg.id != objThisUser.id)
					{
                        var tid = msg.target_id;
						if (aPrivateCounters[tid]) aPrivateCounters[tid]++; else aPrivateCounters[tid] = 1;
						nTotalUnreadCounter++;
						if (!aPrivateMessages[tid]) aPrivateMessages[tid] = [];
						aPrivateMessages[tid].push(msg);
						if (_chat.onPrivateMsgCount) _chat.onPrivateMsgCount(_chat.getUserByID(tid, msg.nick), aPrivateCounters[tid], nTotalUnreadCounter);
					}
				}
			}
		}
	}
	
	/**
	 *	Compares two users.
	 */
	function compareUsers(lval, rval)
	{
		var res = lval.sortKey.localeCompare(rval.sortKey);
		if (res == 0) return rval.nick.localeCompare(lval.nick);
		return res;
	}
	
	/**
	 *	Finds insertion point of user.
	 */
	function findInsertionPoint(user)
	{
		var low = 0, high = aUsers.length, mid = -1;
		while (low < high)
		{
			mid = parseInt((low + high)/2);
			var cmp = compareUsers(aUsers[mid], user);
			if (cmp < 0)
				low = mid + 1;
			else if(cmp > 0)
				high = mid;
			else
				return mid;
		}
		return low;
	}
	
	/**
	 *	Generates a sorting key.
	 */
	function generateSortingKey(user)
	{
		var sortKey;
		switch (user.type)
		{
			case 'moderator':	sortKey = '0'; break;
			case 'owner':		sortKey = '1'; break;
			case 'rj':			sortKey = '2'; break;
			case 'guest':		sortKey = '3'; break;
			case 'user':		sortKey = '4'; break;
			default:			sortKey = '5'; break;
		}
		sortKey += user.nick.toLowerCase();
		return sortKey;
	}
	
	/**
	 *	Adds a user into the sorted list.
	 */
	function addUser(user)
	{
        if (objThisUser && objThisUser.id == user.id) user.current = true;
        user.sortKey = generateSortingKey(user);
        if (user.banned)
        {
            aBanned[user.id] = user;
            if (!isBannedShow()) return;
        }
		var ind = findInsertionPoint(user);
		aUsers.splice(ind, 0, user);
		if (isUserMuted(user.id)) user.muted = true;
		if (_chat.onAddUser) _chat.onAddUser(user, ind, true);
    }
    
    function setUsersList(users)
    {
        /*var t = Date.now();
        console.log('set_user_list, len=', users.length);*/
        /*for (var i in users)
        {
            addUser(users[i]);
        }*/
        var showBanned = isBannedShow();
        for (let i = 0; i < users.length; i++)
        {
            let user = users[i];
            user.sortKey = generateSortingKey(user);
            if (user.banned)
            {
                aBanned[user.id] = user;
                if (!showBanned)
                {
                    users.splice(i, 1);
                    i--;
                }
            }
        }
        users.sort(compareUsers);
        aUsers = users;
        if (_chat.onAddUser)
        {
            for (var i in users)
                _chat.onAddUser(users[i], i, i == users.length - 1);
        }
        /*console.log('set_user_list, time=', Date.now() - t);*/
    }

	/**
	 *	Find an user position in the sorted list.
	 */
	function findUserIndex(user)
	{
		var id = user.id;
		for (var i = 0; i < aUsers.length; i++)
		{
			if (aUsers[i].id == id)
				return i;
		}
		return -1;
	}
	
	/**
	 *	Removes a user from users list.
	 */
	function removeUser(user)
	{
        var ind = findUserIndex(user);
		if (ind !== -1)
		{
			user = aUsers[ind]; 
			aUsers.splice(ind, 1);
			if (_chat.onRemoveUser) _chat.onRemoveUser(user, ind);
		}
	}
	
	/**
	 *	Updates current user. 
	 */
	function updateThisUser(user)
	{
        var prevShowed = isBannedShow();
		var ban = objThisUser !== null ? objThisUser.banned : false; 
		objThisUser = user;
        objThisUser.current = true;
		if (ban != objThisUser.banned && _chat.onBan)
            _chat.onBan(objThisUser.banned);
        if (prevShowed != isBannedShow()) showBanned(!prevShowed);
	}
	
	/**
	 *	Updates a specified user in the users list.
	 */
	function updateUser(user)
	{
		if (objThisUser && user.id == objThisUser.id) updateThisUser(user);
        
        if (user.banned)
        {
            aBanned[user.id] = user;
            if (!isBannedShow()) 
            {
                removeUser(user);
                return;
            }
        }
        else
        {
            delete aBanned[user.id];
        }
        
		var oldPos = findUserIndex(user);
		if (oldPos !== -1)
		{	
			aUsers[oldPos] = user;
			user.sortKey = generateSortingKey(user);
			
			var newPos = findInsertionPoint(user);
			if (newPos != oldPos)
			{
				/* change position*/
				aUsers.splice(oldPos, 1);
				aUsers.splice(newPos, 0, user);
			}
			if (_chat.onUpdateUser) _chat.onUpdateUser(user, oldPos, newPos);
		}
		else
		{
			addUser(user);
		}
	}
	
	/**
	 *	Handler of connection/disconnection.
	 *	@param connected	Boolean flag.
	 */
	function onWebscoketConnect(connected)
	{
		if (connected)
		{
			nTotalUnreadCounter = 0;
            aUsers = [];
            aBanned = {};
            objThisUser = null;
			_connection.send(JSON.stringify({cmd:'init_connection'}));
		}
		else
		{
			if (_chat.onConnect) _chat.onConnect(false);
		}
	}
	
	/**
	 *	Processes server commands.
	 *	@param msg	A message object, received from the server.
	 */
	function onWebsocketMessage(cmd)
	{
		switch (cmd.cmd)
		{
			case 'initialize':
				sSalt = cmd.salt;
				sAppID = typeof(cmd.app_id) != 'undefined' ? cmd.app_id : 0;
				var sHidden = getItem('vchat_hidden_msg_' + sAppID);
				aHiddenMessages = sHidden ? sHidden.split(',').map(function(item) {return parseInt(item, 10);}) : [];
				if (_chat.onConnect) _chat.onConnect(true);
			break;
			
			case 'set_users_list':
                setUsersList(cmd.users);
			break;
			
			case 'add_user':
				addUser(cmd.user);
			break;
			
			case 'remove_user':
                delete aBanned[cmd.user.id];
				removeUser(cmd.user);
			break;
			
			case 'update_user':
				updateUser(cmd.user);
			break;
			
			case 'check_ban':
				if (_chat.onCheckBan) _chat.onCheckBan(cmd.id, cmd.ban);
			break;
			
			case 'send_report':
				if (cmd.type == 'msg')
				{
					if (_chat.onMsgReportSent) _chat.onMsgReportSent(cmd.id, cmd.res);
				}
			break;
			
			case 'set_history':
				if (_chat.onClearHistory) _chat.onClearHistory();
				if (_chat.onAddMessage)
				{
					for (var i in cmd.messages)
					{
						if (!isMessageHidden(cmd.messages[i])) _chat.onAddMessage(cmd.messages[i]);
					}
				}
			break;
			
			case 'clear_history':
				if (_chat.onClearHistory) _chat.onClearHistory();
            break;
            
            case 'accept_msg':
                if (_chat.onMsgSent) _chat.onMsgSent(true);
            break;

            case 'reject_msg':
                if (_chat.onMsgSent) _chat.onMsgSent(false);
            break;
			
			case 'add_msg':
                addMessage(cmd.msg);
			break;

			case 'hide_message':
				_chat.hideMessage(cmd.id);
			case 'delete_message':
				if (_chat.onDeleteMessage) _chat.onDeleteMessage(cmd.id, cmd.ind);
			break;
			
			case 'login':
				if (typeof(cmd.guid) !== 'undefined' && cmd.guid != sGUID)
					setItem('vchat_guid', cmd.guid);
                updateThisUser(cmd.user);
                var usr = _chat.getUserByID(cmd.user.id);
                if (usr) usr.current = true;
				if (_chat.onLogin) _chat.onLogin(cmd.user, findUserIndex(cmd.user));
				sNick = cmd.user.nick;
				if (_chat.onSetNick) _chat.onSetNick(cmd.user.nick);
			break;
			
			case 'logout':
                objThisUser = null;
                showBanned(false);
				if (_chat.onLogout) _chat.onLogout();
				if (_chat.onPoll) _chat.onPoll(objPoll);
			break;
			
			case 'set_nick':
				if (cmd.nick != sNick)
				{
					sNick = cmd.nick;
					setItem('vchat_nick', sNick);
				}
				if (_chat.onSetNick) _chat.onSetNick(cmd.nick);
			break;
			
			case 'set_poll':
				if (typeof(cmd.poll) == 'object')
				{ 
					objPoll = cmd.poll;
					objPoll.answer_index = getPollResult(objPoll.id);
					bShowPollResults = objPoll.show_results ? true : false;
				} 
				else objPoll = cmd.allowed ? null : false;
				if (_chat.onPoll) _chat.onPoll(objPoll);
			break;
			
			case 'set_poll_results':
				if (typeof(cmd.results) == 'object')
				{
					objPoll.results = cmd.results;
					if (_chat.onPollResults) _chat.onPollResults(cmd.results);
				}
			break;
			
			case 'set_poll_params':
				if (typeof(cmd.show_results) != 'undefined')
				{
					bShowPollResults = cmd.show_results ? true : false;
					if (objPoll) objPoll.show_results = bShowPollResults;
					if (_chat.onPollParams) _chat.onPollParams('show_results', bShowPollResults);
				}
            break;
            
            case 'add_private_history':
                if (_chat.onAddMessage)
				{
					for (var i in cmd.messages)
					{
                        var msg = cmd.messages[i];
                        msg.private = true;
                        msg.target_id = cmd.target_id;
						if (!isMessageHidden(msg)) _chat.onAddMessage(msg, true);
					}
				}
            break;

            case 'set_unauth_login':
                if (_chat.onSetUnauthLogin) _chat.onSetUnauthLogin(cmd.value);
            break;
		}
	}
	
	/**
	 *	Set item to the local storage.
	 */
	function setItem(name, value)
	{
		if (typeof(localStorage) !== "undefined")
		{
			try{ localStorage.setItem(name, value); }
			catch(e){ localStorage.clear(); }
		}
		else
		{
			var expires = new Date();
			expires.setDate(expires.getDate() + 365);
			document.cookie =  name + '=' + encodeURIComponent(value) + ";expires=" + expires.toUTCString() +";path=/";
		}
	}
	
	/**
	 *	Get item from the local storage.
	 */
	function getItem(name)
	{
		if (typeof(localStorage) !== "undefined")
		{
			return localStorage.getItem(name);
		}
		else
		{
			var cookie = document.cookie.match(new RegExp("(?:^|; )" + 
				name.replace(/([\.$?*|{}\(\)\[\]\\\/\+^])/g, '\\$1') + "=([^;]*)"));
			return cookie ? decodeURIComponent(cookie[1]) : null;
		}
    }
    
	/**
	 *	Returns an array of polls results. 
	 */
	function getPollResults()
	{
		var sResults = getItem('vchat_poll_results');
		if (!sResults) return [];
		var aResults;
		try { aResults = JSON.parse(sResults); }
		catch(error) { return []; }
		return aResults;
    }
    
	/**
	 *	Returns answer index by poll ID. 
	 */
	function getPollResult(pollID)
	{
		for (var i = 0; i < aPollResults.length; i++)
			if (aPollResults[i].id == pollID) return aPollResults[i].index;
		return -1;
	}
	
	/*
	 *	PUBLIC
	 */
	 
	/* events*/
	this.onConnect				= null;
	this.onLogin				= null;
	this.onLogout				= null;
	this.onSetNick				= null;
    this.onAddMessage			= null;
    this.onMsgSent	        	= null;
	this.onClearHistory			= null;
	this.onDeleteMessage		= null;
	this.onAddUser				= null;
	this.onRemoveUser			= null;
	this.onUpdateUser			= null;
	this.onBan					= null;
	this.onCheckBan				= null;
	this.onMsgReportSent		= null;
	this.onPrivateMsgCount		= null;
	this.onPoll					= null;
	this.onPollResults			= null;
    this.onPollParams			= null;
    this.onSetUnauthLogin       = null;
	
	/**
	 *	Starts connection.
	 */
	this.connect = function(url) 
	{
		_connection.connect(url);
	};
	
	/**
	 *	Closes a connection.
	 */
	this.disconnect = function()
	{
		_connection.disconnect();
	};
	
	/**
	 *	Sends login command to the server.
	 */
	this.login = function(login, password)
	{
		var cmd = {cmd:'login', login:(login || ''), password:(password || '')};
		if (sGUID) cmd.guid = sGUID;
		if (sNick) cmd.nick = sNick;
		if (login && login.indexOf('moderator') == '0' && !bModeratorVisible) cmd.visible = false; 
		_connection.send(JSON.stringify(cmd));
	};
	
	/**
	 *	Salt accessor. 
	 */
	this.getSalt = function() { return sSalt; };
	
	/**
	 *	Send a new message.
	 */
	this.sendMessage = function(text, target_id)
	{
        var localId = getnerateLocalId();
		var cmd = {cmd:'send_message', text:text, local_id: localId };
		if (target_id)
		{
			cmd.target_id = target_id;
		}
		return _connection.send(JSON.stringify(cmd)) ? localId : false;
	};
	
	/**
	 *	Send nick.
	 */
	this.setNick = function(nick)
	{
		if (nick == sNick) return true;
		return _connection.send(JSON.stringify({cmd:'set_nick', nick:nick}));
	};
	
	/**
	 *	Returns current user nick. 
	 */
	this.getNick = function() { return sNick; };
	
    this.getUserID = function() { return objThisUser ? objThisUser.id : ''; };
    
    this.getUser = function() { 
        if (!objThisUser) return null;
        if (Object.assign)
        {
            return Object.assign({}, objThisUser); 
        }
        else
        {
            return {
                id: objThisUser.id,
                type: objThisUser.type,
                nick: objThisUser.nick,
                banned: objThisUser.banned,
                current: objThisUser.current
            }
        }
    };
	
	/**
	 *	Checks if current user can ban other users. 
	 */
	this.isBanAllowed = function()
	{
		if (objThisUser == null) return false;
		return objThisUser.type == "owner" || objThisUser.type == "rj" || objThisUser.type == "moderator";
	};
	
	/**
	 *	Checks if a specified user can be banned. 
	 */
	this.isUserBan = function(userOrType)
	{
		var type = typeof(userOrType) == 'string' ? userOrType : userOrType.type;
		return type != 'moderator' && type != 'owner' && type != 'rj';
	};
	
	/**
	 *	Checks if a specified user can be banned. 
	 */
	this.isCanEditPoll = function()
	{
		if (objThisUser == null) return false;
		return objThisUser.type == "owner" || objThisUser.type == "rj" || objThisUser.type == "moderator";
	};
	
	/**
	 * Checks poll started.  
	 */
	this.isPollStarted = function()
	{
		return !!objPoll;
	};
	
	/**
	 * Returns current status of the poll. 
	 */
	this.getPollStatus = function()
	{
		return objPoll ? objPoll.status : '';
	};
	
	/**
	 * Returns maximal number of poll answers.
	 */
	this.getMaxPollAnswers = function() { return 8; };
	
	/**
	 * Returns answer index for current user.
	 */
	this.getPollAnswerIndex = function()
	{
		return objPoll && typeof(objPoll.answer_index) != 'undefined' ? objPoll.answer_index : -1;
	}; 
	
	/**
	 * Starts a new poll. 
	 */
	this.startPoll = function(questionText, aAnswers)
	{
		return _connection.send(JSON.stringify({cmd:'start_poll', poll:{question:questionText, answers:aAnswers, show_results:bShowPollResults}}));
	};
	
	/**
	 * Stops current poll. 
	 */
	this.stopPoll = function()
	{
		return _connection.send(JSON.stringify({cmd:'stop_poll'}));
	};
	
	/**
	 * Adds a new poll answer. 
	 */
	this.answerPoll = function(answerIndex)
	{
		if (!objPoll) return false;
		if (_chat.getPollAnswerIndex() != -1) return false;
		objPoll.answer_index = answerIndex;
		aPollResults.push({id: objPoll.id, index:  answerIndex});
		/* delete old results*/
		while (aPollResults.length > 100)
			aPollResults.shift();
		setItem('vchat_poll_results', JSON.stringify(aPollResults));
		return _connection.send(JSON.stringify({cmd:'answer_poll', answer_index:answerIndex}));
	};
	
	/**
	 *	Sets a value of the "show results" option.
	 */
	this.setPollShowResults = function(value)
	{
		bShowPollResults = value? true : false;
		if (objPoll)
			return _connection.send(JSON.stringify({cmd:'set_poll_params', show_results:bShowPollResults}));
		else
			return true;
	};
	
	/**
	 *	Returns "show results" option value. 
	 */
	this.isPollShowResults = function() { return bShowPollResults; };
	
	/**
	 *	Returns user object by user ID. 
	 */
	this.getUserByID = function(userID, nick)
	{
		for (var i = 0; i < aUsers.length; i++)
		{
			if (aUsers[i].id == userID)
				return aUsers[i];
        }
        if (nick)
        {
            return {id:userID, nick:nick};
        }
		return null;
    };

    this.getUserByInd = function(ind) { return aUsers[ind]; };
    
    this.usersCount = function() { return aUsers.length; };
    
	
	/**
	 * Set ban for specified user.  
	 */
	this.setUserBan = function(userID, action)
	{
		var ban = action != 'unban'; 
		var cmd = {cmd:'set_ban', ban:ban, user_id:userID};
		if (ban) cmd.duration = action;
		return _connection.send(JSON.stringify(cmd));
	};
	
	/**
	 * Set ban for specified user.  
	 */
	this.setUserMute = function(userID, muted)
	{
		if (muted)
		{
			aMutedUsers.push(userID);
			var maxSize = 100;
			if (aMutedUsers.length > maxSize) aMutedUsers.splice(0, aMutedUsers.length - maxSize);
		}
		else
		{
			var ind = aMutedUsers.indexOf(userID);
			if (ind != -1) aMutedUsers.splice(ind, 1);
		}
		
		var user = _chat.getUserByID(userID);
		if (user)
		{
			user.muted = muted;
			var ind = findUserIndex(user);
			if (_chat.onUpdateUser) _chat.onUpdateUser(user, ind, ind);
		}
		
		setItem('vchat_muted_users', aMutedUsers.join(','));
	};
	
	/**
	 *	 Returns maximal number of messages
	 */
	this.getMaxHistorySize = function() { return nMaxHistorySize; };
	
	/**
	 *	Dispatches addMessages events for all unread private messages of the specified user. 
	 */
	this.getMessages = function(userID, isFirstLoad)
	{
        var aMessages = aPrivateMessages[userID];
        if (isFirstLoad)
        {
            var lastMsgID = aMessages && aMessages.length ? aMessages[0].id : 0;
            _connection.send(JSON.stringify({cmd:'load_private_history', target_id:userID, last_msg_id:lastMsgID }));
        }

        if (!aMessages) return;
        var nick;
		if (_chat.onAddMessage)
		{
			var i;
			for (i in aMessages)
			{
				try	{ if (!isMessageHidden(aMessages[i])) _chat.onAddMessage(aMessages[i]);	}
                catch(e) { console.log(e); }
                nick = aMessages[i].nick;
			}
		}
		delete aPrivateMessages[userID];
		if (aPrivateCounters[userID]) nTotalUnreadCounter -= aPrivateCounters[userID];
		delete aPrivateCounters[userID];
		if (_chat.onPrivateMsgCount) _chat.onPrivateMsgCount(_chat.getUserByID(userID, nick), 0, nTotalUnreadCounter);
	};
	
	/**
	 *	Clears history of common (public) messages. 
	 */
	this.clearHistory = function()
	{
		return _connection.send(JSON.stringify({cmd:'clear_history'}));
	};
	
	/**
	 *	Deletes a message by id or index. 
	 */
	this.deleteMessage = function(id)
	{
		return _connection.send(JSON.stringify({cmd:'delete_message', id:(id? id : 0)}));
	};
	
	/**
	 *	Hides a messsage for current user. 
	 */
	this.hideMessage = function(id)
	{
        if (_chat.onDeleteMessage) _chat.onDeleteMessage(id);
        aHiddenMessages.push(parseInt(id,10));
        var maxSize = 50;
        if (aHiddenMessages.length > maxSize) aHiddenMessages.splice(0, aHiddenMessages.length - maxSize);
        setItem('vchat_hidden_msg_'+sAppID, aHiddenMessages.join(','));
	};
	
	/** Returns visiblity of banned users. */
	this.isBannedVisible = function() { return bBannedVisible; };
	
	/**
	 *	Stores banned visibility options. 
	 */
	this.setBannedVisible = function(bVisible) 
	{
        var prevShowed = isBannedShow();
		bBannedVisible = bVisible;
        setItem('vchat_show_banned', bVisible ? '' : 'false');
        if (prevShowed != isBannedShow()) showBanned(!prevShowed);
	};
	
	/**
	 *	Reads moderator visibility option.
	 */
	this.isModeratorVisible = function()
	{
		return bModeratorVisible;
	};
	
	/**
	 *	Sets moderator visibility option.
	 */
	this.setModeratorVisible = function(bVisible)
	{
		bModeratorVisible = bVisible ? true : false;
		return _connection.send(JSON.stringify({cmd:'set_user_visibility', visible:bModeratorVisible}));
	};
	
	/**
	 * Check ban status by message id or index. 
	 */
	this.checkBan = function(msgId, msgIndex)
	{
		return _connection.send(JSON.stringify({cmd:'check_ban', id:(msgId? msgId : 0), ind:(msgIndex? msgIndex: 0)}));
	};
	
	/**
	 * Sends message report. 
	 */
	this.sendMsgReport = function(msgId, value)
	{
		return _connection.send(JSON.stringify({cmd:'send_report', id:(msgId? msgId : 0), type:'msg', value:value}));
    };
    
    /**
     * Set unauth login
     */
    this.setUnauthLogin = function(unauthAllowed)
    {
        return _connection.send(JSON.stringify({cmd:'set_unauth_login', value:(!!unauthAllowed)}));
    };
}
