/**
 *	Scroll bar class. 
 */
function CScrollBar(container, hide)
{
    var scrollbar = null,
        contentPosition = 0,
	    scrollerBeingDragged = false,
	    scroller,
	    topPosition,
        cltHeight = 0;
    var minScrollerHeight = 20;
    var updateTime = 0;
    var nUpdateTimeout = false;
        
    function getScrollerHeight()
    {
        var h = container.scrollHeight;
		if (h == 0) h = 1;
	    var visibleRatio = container.clientHeight / h;
	    return Math.max(visibleRatio * container.clientHeight, minScrollerHeight);
    }
	
	function updateScrollerHeight()
	{
	    var height = getScrollerHeight();
	    scroller.style.height = height + 'px';
        scroller.style.display =  container.clientHeight < container.scrollHeight ? 'block' : 'none';
        if (hide)
        {
            scrollbar.style.display = scroller.style.display;
        }
	}
	
	function moveScroller(evt)
	{
	    /* Move Scroll bar to top offset*/
        updateScrollerTop(evt.target);
    }
    
    function updateScrollerTop(cont)
    {
        cltHeight = container.clientHeight;
        var h = container.scrollHeight - cltHeight;
        if (h <= 0) h = 1;
        var scrollPercentage = cont.scrollTop / h;
        topPosition = scrollPercentage * (cltHeight - getScrollerHeight());
        scroller.style.top = topPosition + 'px';
    }
	
	function startDrag(evt)
	{
	    normalizedPosition = evt.pageY;
	    contentPosition = container.scrollTop;
	    scrollerBeingDragged = true;
	    /* disable selection*/
	     if(evt.stopPropagation) evt.stopPropagation();
	    if(evt.preventDefault) evt.preventDefault();
	    evt.cancelBubble=true;
	    evt.returnValue=false;
	    return false;
	}
	
	function stopDrag(evt)
	{
	    scrollerBeingDragged = false;
	}
	
	function scrollBarScroll(evt)
	{
	    if (scrollerBeingDragged === true)
	    {
            if (container.scrollHeight > container.clientHeight && container.clientHeight > getScrollerHeight())
            {
                var dif = evt.pageY - normalizedPosition;
                var scrollEquivalent = dif * ((container.scrollHeight - container.clientHeight) / (container.clientHeight - getScrollerHeight()));
                container.scrollTop = contentPosition + scrollEquivalent;
            }
	    }
	}
	
	function createScroller()
	{
		/* Creates scroller element and appends to scrollable div*/
		/* create scroller element*/
		scrollbar = document.createElement("div");
		scrollbar.className = 'chat_scroll';
		scroller = document.createElement("div");
		scroller.className = 'chat_scroll_slider';
		
		/* determine how big scroller should be based on content*/
		updateScrollerHeight();
		
		/* append scroller to container div*/
		scrollbar.appendChild(scroller);
		container.parentNode.appendChild(scrollbar);
		
		/* attach related draggable listeners*/
		scroller.addEventListener('mousedown', startDrag);
		window.addEventListener('mouseup', stopDrag);
		window.addEventListener('mousemove', scrollBarScroll);
	}
	
	function update()
	{
        if (nUpdateTimeout !== false) return; /* update timer alredy started*/
        /* limit refresh rate*/
        var t = Date.now();
        const timeout = 100;
        if (t - updateTime < timeout)
        {
            nUpdateTimeout = setTimeout(function() {
                nUpdateTimeout = false;
                updateScroll(Date.now());
            }, timeout - (t - updateTime));
            return;
        }

        updateScroll(t);
    }

    function updateScroll(t)
    {
        updateTime = t;
        if (cltHeight > container.clientHeight)
        {
            container.scrollTop += cltHeight - container.clientHeight;
        }
        updateScrollerHeight();
        updateScrollerTop(container);
        container.style['padding-right'] = (nScrollPadding - (container.offsetWidth - container.clientWidth) + nScrollWidth) + 'px';
    }
	
	function isCssVars() { return window.CSS && window.CSS.supports && window.CSS.supports('--fake-var', 0); }
	
	createScroller();
	var style = getComputedStyle(document.documentElement);
	var nScrollPadding = isCssVars() ? parseInt(style.getPropertyValue('--vchat-scroll-padding')) : 60;
	var nScrollWidth = isCssVars() ? parseInt(style.getPropertyValue('--vchat-scroll-width')) : 5;
	/* fix right padding for width*/
	container.style['overflow-y'] = 'scroll';
	container.style['box-sizing'] = 'border-box';
	update();
	/* Listeners */
	container.addEventListener('scroll', moveScroller);
	container.customScrollBar = {update:update};
};

/**
 *	Chat GUI.
 */
function	ChatGUI(domChatContainer, sURL, sUSettingsURL, funcLoginHandler, aSmileys, sBlankImgURL, sMsgBlipSoundURL)
{
	/* variables*/
	var bFirstLogin			= true;
	var nMaxSmileysInMsg	= 5;
	var domChatContent		= domChatContainer.querySelector('.chat_cont');
	var domLanguage			= domChatContent.querySelector('[data-cont="language"]');
	var domHistory			= domChatContent.querySelector('.chat_history[data-tab="common"]');

	var domScrollDown		= domChatContent.querySelector('.chat_scroll_down');
	var domCurrentHistory	= domHistory;
	var domHistoryTabs		= domChatContent.querySelector('.chat_tabs');
	var domPrivateTabs		= domChatContent.querySelector('.chat_private_tabs');
	var domHistoryPages		= domChatContent.querySelector('.chat_history_pages');
	var domUsers		    = domChatContent.querySelector('.chat_users');
	var domPreloader		= domChatContent.querySelector('.chat_preloader');
	var domSmyleTab 		= domChatContent.querySelectorAll('.smile_wrapper');
    var objMsgBlipSound		= null;
    
	var curMessage          = null;
	var countNewMassege     = 0;
	var isFirstLogin 	    = true;
    domPreloader.classList.add('chat_showed');
    var domUsersList = document.createElement('div');
    domUsersList.style.position = 'relative';
    domUsers.appendChild(domUsersList);
    domUsersList.addEventListener('click', onUsersClick);
    domUsers.addEventListener('scroll', onUsersScroll);
		
	/* create chat object*/
	var objChat = new Chat();
	objChat.onConnect		= onConnect;
	objChat.onLogin			= onLogin;
	objChat.onLogout		= onLogout;
    objChat.onSetNick		= onSetNick;
    objChat.onMsgSent       = onMsgSent;
	objChat.onAddMessage	= onAddMessage;
	objChat.onClearHistory	= onClearHistory;
	objChat.onDeleteMessage = onDeleteMessage;
	objChat.onAddUser		= onAddUser;
	objChat.onRemoveUser	= onRemoveUser;
	objChat.onUpdateUser	= onUpdateUser;
	objChat.onBan			= onBan;
	objChat.onCheckBan		= onCheckBan;
	objChat.onMsgReportSent = onMsgReportSent;
    objChat.onPrivateMsgCount = onPrivateMsgCount;

    objChat.onSetUnauthLogin = function (value) {
        if (value)        
            domAllowUnauthBtn.classList.add('chat_checked');
        else
            domAllowUnauthBtn.classList.remove('chat_checked');
        domAllowUnauthBtn.classList.remove('chat_disabled');
    };
	
	if (window.WSChat) window.WSChat.setChat(objChat);
	
    objChat.connect(sURL);

    function setDisplay(element, isDisplay)
    {
        if (isDisplay)
            element.classList.remove('chat_hidden');
        else
            element.classList.add('chat_hidden');
    }
    
    function initHistoryScroll(historyNode)
    {
        var scrollDown = historyNode.parentNode.querySelector('.chat_scroll_down');
        if (!scrollDown) {
            scrollDown = domScrollDown.cloneNode();
            scrollDown.style.right = '-30px';
            historyNode.parentNode.insertBefore(scrollDown, historyNode);
        }
        scrollDown.addEventListener('click',function() {
            historyNode.scrollTop = historyNode.scrollHeight;
            scrollDown.style.right = '-30px';
        });

        historyNode.addEventListener('scroll', function() {
            scrollDown.style.right = (historyNode.scrollTop + historyNode.offsetHeight < historyNode.scrollHeight - historyNode.offsetHeight/2) ? '23px' : '-30px';
            updateTopDate(historyNode);
        });
        CScrollBar(historyNode);
    }

	/* send button*/
	var domSendBtn = domChatContent.querySelector('.chat_send_btn');
	domSendBtn.addEventListener('click', function(){ SendMessage(); });
	
	/* init scrollbars*/
    CScrollBar(domUsers);
    initHistoryScroll(domHistory);
	for(var i =0; i < domSmyleTab.length; i++){
		CScrollBar(domSmyleTab[i]);
	}
	
	
	/* showing left panel (for mobile view)*/
	var domShowLeftBtn = domChatContent.querySelector('.chat_show_users_btn');
	let domLeftPanel = domChatContent.querySelector('.chat_left');
	if (domShowLeftBtn && domLeftPanel) {
		domShowLeftBtn.addEventListener('click', function() {
			domLeftPanel.classList.toggle('chat_showed');
		});
	}
	
	/* private tabs click*/
	domHistoryTabs.addEventListener('click', onHistoryTabsClick);
	domHistoryTabs.querySelector('.chat_scroll_left').addEventListener('click', function(event){
			domPrivateTabs.scrollLeft -= domPrivateTabs.clientWidth/2;
		});
	domHistoryTabs.querySelector('.chat_scroll_right').addEventListener('click', function(event){
			domPrivateTabs.scrollLeft += domPrivateTabs.clientWidth/2;
		});
	
	/* ban dialog*/
	var domBanDlg = initButtonsDlg('.chat_ban_dlg', onBanUserClick);
	var domMuteDlg = initButtonsDlg('.chat_mute_dlg', onBanUserClick);
	
	function initButtonsDlg(selector, clickHandler)
	{
		var dlg = domChatContent.querySelector(selector);
		if (!dlg) return null;
		var buttons = dlg.querySelectorAll('.chat_push_btn');
		for (var i = 0; i < buttons.length; i++)
			buttons[i].addEventListener('click', clickHandler);
		dlg.querySelector('.chat_close_btn').addEventListener('click', function(e){ dlg.classList.remove('chat_showed'); });
		return dlg;
	};
    
    /** 
     * Scroll to specified tab
     */
    function scrollToTab(tab)
    {
        var cont = tab.parentNode;
        if (tab.offsetLeft < cont.scrollLeft) cont.scrollLeft = tab.offsetLeft - 5;
        if (tab.offsetLeft + tab.offsetWidth > cont.scrollLeft + cont.clientWidth) cont.scrollLeft = tab.offsetLeft + tab.offsetWidth - cont.clientWidth + 5;
    }
	
	/* settings dialog*/
	var domSettingsDlg = domChatContent.querySelector('.chat_settings');
	var domHearMsgSoundBtn	= null;
	var domShowBannedBtn	= null;
    var domClearHistoryBtn	= null;
    var domAllowUnauthBtn   = null;
	if (domSettingsDlg)
	{ 
		var domSettingsBtn = domChatContent.querySelector('.chat_settings_btn'); 
		domSettingsBtn.addEventListener('click', function(event){
				domSettingsDlg.classList.toggle('chat_showed');
			});
		domShowBannedBtn = domSettingsDlg.querySelector('.chat_show_banned_btn');
		if (objChat.isBannedVisible()) domShowBannedBtn.classList.add('chat_checked');
		domShowBannedBtn.addEventListener('click', onShowBannedClick);
		/* clear history click*/
		domClearHistoryBtn = domSettingsDlg.querySelector('.chat_clear_history_btn');
		domClearHistoryBtn.addEventListener('click', function(event){
				if (objChat.clearHistory()) domSettingsDlg.classList.remove('chat_showed');
			});
		var domShowModeratorBtn = domSettingsDlg.querySelector('.chat_show_moderator');
		if (domShowModeratorBtn)
		{
			if (objChat.isModeratorVisible()) domShowModeratorBtn.classList.add('chat_checked');
			domShowModeratorBtn.addEventListener('click', onShowModeratorClick);
		}
		domHearMsgSoundBtn = domSettingsDlg.querySelector('.chat_hear_msg_sound');
		domHearMsgSoundBtn.classList.add('chat_checked');
		domHearMsgSoundBtn.addEventListener('click', function(event){
				domHearMsgSoundBtn.classList.toggle('chat_checked');
            });
            
        domAllowUnauthBtn = domSettingsDlg.querySelector('.chat_allow_unauth_login');
        domAllowUnauthBtn.addEventListener('click', function(event){
            objChat.setUnauthLogin(!domAllowUnauthBtn.classList.contains('chat_checked'));
            domAllowUnauthBtn.classList.add('chat_disabled');
        });
	}
	
	/* poll dialog*/
	var domPollDlg = domChatContent.querySelector('.chat_poll');
	var domPollBtn = domChatContent.querySelector('.chat_poll_btn'); 
	domPollBtn.addEventListener('click', function(event){
			/*updatePollDlgPosition();*/
			domPollDlg.classList.toggle('chat_showed');
		});
    var objPoll = new ChatPollGUI();
    CScrollBar(domPollDlg.querySelector('.chat_poll_questions'), true);
	
	/* smileys dialog*/
	var domSmileysDlg = domChatContent.querySelector('.chat_smileys');
	domSmileysDlg.addEventListener('click', function(event){
			var id = event.target.getAttribute('data-id'); 
			if (id) onClickSmile(event.target, id);
		});
	var domSmileysTabs = domChatContent.querySelector('.smileys_tabs');
	domSmileysDlg.querySelector('.chat_scroll_left').addEventListener('click', function(event){
			domSmileysTabs.scrollLeft -= domSmileysTabs.clientWidth/2;
		});
	domSmileysDlg.querySelector('.chat_scroll_right').addEventListener('click', function(event){
			domSmileysTabs.scrollLeft += domSmileysTabs.clientWidth/2;
		});
		
	/* smileys tab click*/
	domSmileysTabs.addEventListener('click', function(event){
			if (event.target != domSmileysTabs) /* tab*/
			{
				var currentTab = domSmileysTabs.querySelector('.selected');
				if (currentTab != event.target)
				{
					currentTab.classList.remove('selected');
                    event.target.classList.add('selected');
                    domSmileysDlg.querySelector('.smileys_sheet.selected').classList.remove('selected');
                    var domSheet = domSmileysDlg.querySelector('.smileys_sheet[data-tab="'+ event.target.getAttribute('data-tab') +'"]');
					domSheet.classList.add('selected');
					domSheet.querySelector('.smile_wrapper').customScrollBar.update();
                }
                scrollToTab(event.target);
			}
		});
	var domSmileBtn = domChatContent.querySelector('.chat_smile_btn'); 
	domSmileBtn.addEventListener('click', function(event){
			var disp = domSmileysDlg.classList.contains('chat_showed');
			domMessageInput.focus();
            domSmileysDlg.classList.toggle('chat_showed');
			if (!disp && domSmileysTabs.scrollWidth > domSmileysTabs.clientWidth) domSmileysTabs.classList.add('scrolled');

			for(var i= 0; i < domSmyleTab.length; i++)
				domSmyleTab[i].customScrollBar.update();
		});
	
	/* inputs*/
	var domNickInput = domChatContent.querySelector('.chat_nick_input');
	domNickInput.addEventListener("keypress", function(event){
		if (event.keyCode == 10 || event.keyCode == 13) { event.preventDefault(); SendNick(); }
		});
	domNickInput.addEventListener("blur", function(event) { SendNick(); }); 
	if (sUSettingsURL)
	{
		domNickInput.addEventListener("click", function(event){
			if (domNickInput.classList.contains("button")) window.open(sUSettingsURL, '_blank');
			});
	}
	var domMessageInput = domChatContent.querySelector('.chat_message_input');
	domMessageInput.addEventListener("keypress", function(event){
		if (event.keyCode == 10 || event.keyCode == 13) { event.preventDefault(); SendMessage(); }
		});
	if ("MutationObserver" in window)
	{
		var objMessageInputObserver = new MutationObserver(onMessageInputChange);
		objMessageInputObserver.observe(domMessageInput, {attributes: false, childList: true, characterData: true, subtree: true});
	}
	else
		domMessageInput.addEventListener('DOMSubtreeModified', onMessageInputChange);
	
	/* hide dialogs*/
	window.addEventListener('mousedown', function(e) {
		if (!isParent(e.target, domSmileysDlg) && e.target != domMessageInput && e.target != domSmileBtn)
			domSmileysDlg.classList.remove('chat_showed');
		if (!isParent(e.target, domSettingsDlg) && e.target != domSettingsBtn)
			domSettingsDlg.classList.remove('chat_showed');
		if (!isParent(e.target, domPollDlg) && e.target != domPollBtn)
			domPollDlg.classList.remove('chat_showed');
		if (!isParent(e.target, domMsgMenu))
			hideMsgMenu();
		if (domLeftPanel && !isParent(e.target, domLeftPanel))
			domLeftPanel.classList.remove('chat_showed');
	});
	
	/* social login*/
	domChatContent.querySelector('.chat_login').addEventListener("click", onLoginClick);
	
    /* history click*/
    var historyCont = domChatContent.querySelector('.chat_history_container');
    var bFullMessages = historyCont.classList.contains('chat_fullmsg');
    
	historyCont.addEventListener("click", function(event){
		if (event.target.classList.contains('chat_menu'))
		{
			if (isPseudoHover(event, '::after')) showMessageMenu(event.target);
		}
		else if (event.target.classList.contains('chat_usr')){ /* user nick click*/
			var text = event.target.textContent;
			insertInputNode(document.createTextNode(text), text);
		}
		if(event.target.classList.value.indexOf("smile_") + 1){
			var node = document.createElement('img');
			node.className = event.target.classList.value;
			let id = event.target.getAttribute('data-id');
        	node.setAttribute('data-id', id);
        	node.src = event.target.getAttribute('src');
			insertInputNode(node, '%' + id + '%');
		}
	});
	/* history click*/
	historyCont.addEventListener("mouseover", function(event){
		if (event.target.classList.contains('chat_msg'))
		{
			if (objChat.getUserID() != '')
			{
				if (!event.target.classList.contains('chat_menu')) event.target.classList.add('chat_menu');
			}
			else
			{
				if (event.target.classList.contains('chat_menu')) event.target.classList.remove('chat_menu');
			}
		}
	});
	/*message menu*/
	var domMsgMenu = domChatContent.querySelector('.chat_msg_menu');
	var msgMenuButtons = domMsgMenu.querySelectorAll('.chat_push_btn');
	for (var i = 0; i < msgMenuButtons.length; i++)
		msgMenuButtons[i].addEventListener('click', onMsgMenuClick);
	var domReportMenu = domChatContent.querySelector('.chat_report_menu');
	if (domReportMenu)
	{
		domReportMenu.querySelector('.chat_close_btn').addEventListener('click', function(e){ domReportMenu.classList.remove('chat_showed'); });
		domReportMenu.addEventListener('click', onReportMenuClick);
    }
    
    /** Date format */

    var monthNames = ['January', 'February', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October', 'November', 'December'];
    var dateFormat = {full:'%m %d, %y', short:'%m %d', today:'today', yesterday:'yesterday'};
    if (domLanguage) {
        var node = domLanguage.querySelector('[data-id=chat_month_names');
        if (node) monthNames = node.innerText.split(',');
        node = domLanguage.querySelector('[data-id=date_format');
        if (node) 
        {
            try { dateFormat = JSON.parse(node.innerText); }
            catch(error) { console.log(error); 
            }
        }
    }
        
    /** Format date to display */
    function getDateHTML(date)
    {
        var year = date.getFullYear();
        var month = date.getMonth();
        var day = date.getDate();

        var format = getDateFormat(year, month, day);
        
        var dateKey = (format == 'today' || format == 'yesterday') ?
            ' data-date="'+ year +'-'+ month +'-'+ day +'"'
            : '';

        return '<span'+ dateKey +'>' + getDateStr(format, year, month, day) + '</span>';
    }

    /** Genarates date string */
    function getDateStr(format, year, month, day)
    {
        return (dateFormat[format] || dateFormat.short).replace('%m', monthNames[month]).replace('%d', day).replace('%y', year);
    }

    /** Get format name for specified date */
    function getDateFormat(year, month, day)
    {
        var now = new Date();

        if (year == now.getFullYear())
        {
            if (month == now.getMonth() && day == now.getDate())
            {
                return 'today';
            }
            else  
            {
                now.setDate(now.getDate() - 1);
                if (month == now.getMonth() && day == now.getDate())
                {
                    return 'yesterday';
                }
            }
            
            return 'short';
        }

        return 'full';
    }

    /* Check date change by timer*/

    var checkDateDay = (new Date()).getDate();
    setInterval(function() {
        var now = new Date();
        if (now.getDate() != checkDateDay)
        {
            checkDateDay = now.getDate();
            var dates = domChatContent.querySelectorAll('.chat_msg_date [data-date]');
            for (var i = 0; i < dates.length; i++)
            {
                var node = dates[i];
                var aDate = node.getAttribute('data-date').split('-');
                var format = getDateFormat(aDate[0], aDate[1], aDate[2]);
                node.innerText = getDateStr(format, aDate[0], aDate[1], aDate[2]); /* update date format: today -> yesterday, yesterday -> date number*/
            }
        }
    }, 1000);
    
    /** Helper function. Insert node after specified node */
    function insertAfter(newNode, referenceNode) {
        referenceNode.parentNode.insertBefore(newNode, referenceNode.nextSibling);
    }
		
	/**
	 *	Detects mouse hover on pseudo element.
	 * 
	 *	@param {Object} event	The mouse event object.
	 *	@param {String} pseudo	Name of the pseudo element (ex: '::before', '::after'...).
	 *	@param {bool} skip		If it is true, then the function returns false if pseudo element not found.  	
	 */ 
	function isPseudoHover(event, pseudo, skip)
	{
		var x = event.offsetX;
		var y = event.offsetY;
		var style = getComputedStyle(event.target, pseudo);
		if (typeof(style.content) != 'string' || style.content == '' || style.content == 'none') return skip ? false : true;
		var l = parseInt(style.left);
		var t = parseInt(style.top);
		if (isNaN(l) || isNaN(t)) return skip ? false : true;
		var w = parseInt(style.width);
		var h = parseInt(style.height);
		return (x >= l && x < (l+w) && y >= t && y < (t+h)) ? true : false;
	}
	
	/**
	 * Shows message menu. 
	 * @param {Object} target
	 * @param {string} userID
	 */
	function showMessageMenu(target)
	{
		/* init buttons*/
		if(target.classList.contains('chat_ut_priv') || target.classList.contains('chat_ut_my')) return;

		var uid = target.getAttribute('data-uid').split('|');
		var bMyMsg = uid[1] == objChat.getUserID();
		var bDelete = objChat.isBanAllowed() || bMyMsg;
		for (var i = 0; i < msgMenuButtons.length; i++)
		{
			var visible = false;
			switch (msgMenuButtons[i].getAttribute('data-action'))
			{
                case 'private_msg': visible = !bMyMsg; break;
				case 'delete':      visible = bDelete; break;
				case 'hide':        visible = !bDelete; break;
                case 'report':      visible = !bMyMsg && !bDelete; break;
			}
			msgMenuButtons[i].style.display = visible ? 'block' : 'none';
		}
		
		if (objChat.isBanAllowed() && objChat.isUserBan(uid[0]))
			objChat.checkBan(target.getAttribute('data-id'), target.getAttribute('data-ind'));
		
		/* calculate dialog position*/
		var trect = target.getBoundingClientRect();
		var prect = domMsgMenu.parentNode.getBoundingClientRect();
		var left = trect.left - prect.left + target.offsetWidth - domMsgMenu.offsetWidth;
		var top = trect.top - prect.top;
		domMsgMenu.classList.add('chat_showed');
		if (top + domMsgMenu.offsetHeight > domMsgMenu.parentNode.offsetHeight)
			top = domMsgMenu.parentNode.offsetHeight - domMsgMenu.offsetHeight;
		domMsgMenu.style.left = left + "px";
		domMsgMenu.style.top = top + "px";
		target.classList.add('chat_selected');
	}
	
	/**
	 *	Handler of ban checking event. 
	 */
	function onCheckBan(id, banned)
	{
		var sel = domChatContent.querySelector('.chat_selected');
		if (sel && sel.getAttribute('data-id') == id)
		{
			domMsgMenu.querySelector('.chat_push_btn[data-action="'+(!banned ? 'ban' : 'unban')+'"]').style.display = 'block';
			var top = domMsgMenu.offsetTop;
			if (top + domMsgMenu.offsetHeight > domMsgMenu.parentNode.offsetHeight)
			{
				top = domMsgMenu.parentNode.offsetHeight - domMsgMenu.offsetHeight;
				domMsgMenu.style.top = top + "px";
			}
		}
	}
	
	/**
	 *	Handler of a click event on a button of the message menu.
	 */
	function onMsgMenuClick(event)
	{
		var sel = domChatContent.querySelector('.chat_selected');
		if (sel)
		{
			switch(event.target.getAttribute('data-action'))
			{
                case 'private_msg':
                    var userID = sel.getAttribute('data-uid').split('|')[1];
                    var nick = getUserNickFromMessage(sel) || userID;
                    if (getPrivateHistoryTab(objChat.getUserByID(userID, nick)))
                        showHistoryPage(userID); 
                break;
				case 'delete': objChat.deleteMessage(sel.getAttribute('data-id')); break;
				case 'hide': objChat.hideMessage(sel.getAttribute('data-id')); break;
				case 'report':
					domReportMenu.classList.add('chat_showed');
					domReportMenu.classList.remove('chat_disabled');
					domReportMenu.setAttribute('data-id', sel.getAttribute('data-id'));
					break;
				case 'ban': showBanDlg(event.target.parentNode, sel.getAttribute('data-uid').split('|')[1], sel); break;
				case 'unban': objChat.setUserBan(sel.getAttribute('data-uid').split('|')[1], 'unban'); break;
			}
		}
		hideMsgMenu();
	}
	
	/**
	 * Hides messge menu. 
	 */
	function hideMsgMenu()
	{
		domMsgMenu.classList.remove('chat_showed');
		var sel = domChatContent.querySelector('.chat_selected');
		if (sel) sel.classList.remove('chat_selected');
	}
	
	/**
	 * Click on report menu. 
	 */
	function onReportMenuClick(event)
	{
		if (event.target.classList.contains('chat_push_btn'))
		{
			objChat.sendMsgReport(domReportMenu.getAttribute('data-id'), event.target.getAttribute('data-type'));
			domReportMenu.classList.add('chat_disabled');
		}
	}
	
	/**
	 * Handler of event of report sending.  
	 */
	function onMsgReportSent(id, res)
	{
		if (domReportMenu.getAttribute('data-id') == id)
		{
			domReportMenu.classList.add('chat_hide');
			domReportMenu.classList.remove('chat_showed');
			var tid = setTimeout(function(){
				domReportMenu.classList.remove('chat_hide');
				},
				1000);
		}
	}
	
	/**
	 *	Inserts a node into message input. 
	 */
	function insertInputNode(node, fallbackText)
	{
		domMessageInput.focus();
		var sel, range;
		if (window.getSelection)
		{
		    sel = window.getSelection();
		    if (sel.getRangeAt && sel.rangeCount) {
		        range = sel.getRangeAt(0);
		        range.deleteContents();
		        range.insertNode(node);
		        range = document.createRange();
		        range.setStartAfter(node);
             	range.setStartAfter(node);
             	sel.removeAllRanges();
             	sel.addRange(range);
		    }
		}
		else if (document.selection && document.selection.createRange)
		{
		    document.selection.createRange().text = text;
		}
	}
	
	/**
	 *	Updates chat visibility. 
	 */
	this.updateVisibility = function()
	{
		domUsers.customScrollBar.update();
        domHistory.customScrollBar.update();
        domHistory.scrollTop = domHistory.scrollHeight;
        onUsersScroll();
		for(var i =0; i< domSmyleTab.length; i++){
			domSmyleTab[i].customScrollBar.update();
		}
		
	};
	
	/**
	 *	Checks if element 'el' has specified parent.  
	 */
	function isParent(el, parent)
	{
		while (el)
		{
			if (el == parent) return true;
			el = el.parentNode;
		}
		return false;
	}
	
	/**
	 *	Returns private tab for specified user.
	 * 
	 * @param {Object} item		User or message object.	
	 */
	function getPrivateHistoryTab(item)
	{
		if (!item) return;
		
		var domTab = domPrivateTabs.querySelector('[data-tab="'+item.id+'"]');
		if (!domTab)
		{
			/* create history page*/
			var domPageContainer = document.createElement('div');
			domPageContainer.className = 'chat_history_wrapper';
			var domPage = document.createElement('div');
			domPage.className = 'chat_history';
			domPage.setAttribute('data-tab', item.id);
			
			/* create tab*/
			domTab = document.createElement('div');
			var domNick = document.createElement('span');
			domNick.innerText = item.nick;
			domTab.appendChild(domNick);
			var domCounter = document.createElement('span');
			domCounter.className = 'counter';
			domCounter.style.display = 'none';
			domTab.appendChild(domCounter);
			domTab.setAttribute('data-tab', item.id);
			var domClose = document.createElement('span');
			domClose.className = 'chat_close_tab';
			domTab.appendChild(domClose);
			var numTabs = domPrivateTabs.children.length; 
			
			/* add to DOM*/
			domPageContainer.appendChild(domPage);
			domHistoryPages.appendChild(domPageContainer);
            domPrivateTabs.appendChild(domTab);
            initHistoryScroll(domPage);
			if (domPrivateTabs.scrollWidth > domPrivateTabs.clientWidth) domPrivateTabs.classList.add('scrolled');
			
			/* show tabs*/
			if (numTabs == 0) domHistoryPages.classList.add('tabs');
		}
		
		return domTab;
	}
	
	/**
	 *	Handler of click event on history tabs. 
	 */
	function onHistoryTabsClick(event)
	{
		var tabID = event.target.getAttribute('data-tab');
		
		if (tabID)
			showHistoryPage(tabID, event.target);
		else if (event.target.classList.contains('chat_close_tab'))
		{
			tabID = event.target.parentElement.getAttribute('data-tab');
            closeHistoryPage(tabID, event.target.parentElement);
		}
	}
	
	/**
	 *	Show a history page by the tab ID.
	 * 
	 *	@param {string} tabID	ID of the page.
	 *	@param {Object} domTab	Optional, DOM element of page tab.
	 */
	function showHistoryPage(tabID, domTab)
	{
		var currentTab = domHistoryTabs.querySelector('.selected');	

		if(tabID == "common"){
			updateCounterHistory(false);
		}

		if (!domTab) domTab = domHistoryTabs.querySelector('[data-tab="'+ tabID +'"]');
		if (domTab)
		{
            if (currentTab != domTab)
            {
                currentTab.classList.remove('selected');
                domTab.classList.add('selected');
                domHistoryPages.querySelector('.chat_history_wrapper.selected').classList.remove('selected');
                domCurrentHistory = domHistoryPages.querySelector('.chat_history[data-tab="'+ tabID +'"]');
                domCurrentHistory.parentNode.classList.add('selected');
                domCurrentHistory.customScrollBar.update();
                if (tabID != 'common')
                {
                    var isFirstOpen = domCurrentHistory.hasAttribute('data-get') == false;
                    objChat.getMessages(tabID, isFirstOpen);
                    if (isFirstOpen)
                        domCurrentHistory.setAttribute('data-get', '1');
				}			
			}
            scrollToTab(domTab);
        }
	}
	
	/**
	 *	Close a history page by the tab ID.
	 * 
	 *	@param {string} tabID	ID of the page.
	 *	@param {Object} domTab	Optional, DOM element of page tab.
	 */
	function closeHistoryPage(tabID, domTab)
	{
		if (!domTab) domTab = domHistoryTabs.querySelector('[data-tab="'+ tabID +'"]');
		if (domTab)
		{
			var currentTab = domHistoryTabs.querySelector('.selected');
			if (currentTab == domTab)
				showHistoryPage('common');
			
			domTab.parentNode.removeChild(domTab);
			var page = domHistoryPages.querySelector('.chat_history[data-tab="'+ tabID +'"]');
            if (page) domHistoryPages.removeChild(page.parentNode);

            
            if (domPrivateTabs.scrollWidth <= domPrivateTabs.clientWidth)
                domPrivateTabs.classList.remove('scrolled');
            if (domPrivateTabs.childNodes.length == 0) 
            {
                domHistoryPages.classList.remove('tabs');
                domHistory.customScrollBar.update();
            }
		}
	}

	/** Connect/disconnect event handler. */
	function onConnect(connected)
	{
		if (connected)
		{
			domUsersList.innerHTML = "";
			domPreloader.classList.remove('chat_showed');
			/* login*/
			if (funcLoginHandler)
			{
				funcLoginHandler(objChat.getSalt(),	function(data) {
					var res;
					try	{ res = data; }
					catch (error)
					{
						console.log(error);
						objChat.login();
						return;
					}
					if (res.result == 'OK')
						objChat.login(res.login, res.password);
					else
						objChat.login();
				});
			}
			else
				objChat.login();
		}
		else
		{
			domPreloader.classList.add('chat_showed');
			domSendBtn.classList.add('chat_disabled');
			domSmileBtn.classList.add('chat_disabled');
			domPollBtn.classList.add('chat_disabled');
			domSettingsBtn.classList.add('chat_disabled');
			bFirstLogin = true;
		}
	}

	/**
	 * Login event handler.
	 * @param {Object} user		User object.
	 * @param {Object} index	Index of the user item in the users list.
	 */
	function onLogin(user, index)
	{
		domChatContent.querySelector('.chat_login').classList.remove('chat_showed');
		domMessageInput.setAttribute('contenteditable', 'true');
		if (user.type == "simple" || user.type == "guest" || user.type == "moderator")
		{
			domNickInput.setAttribute('contenteditable', 'true');
			domNickInput.classList.remove('button');
		}
		else
		{
			domNickInput.classList.add('button');
			domNickInput.setAttribute('contenteditable', 'false');
        }
        
        domChatContent.classList.add('chat_logged_in');
		
		/* enable button*/
		domSendBtn.classList.remove('chat_disabled');
		domSmileBtn.classList.remove('chat_disabled');
		domPollBtn.classList.remove('chat_disabled');
		domSettingsBtn.classList.remove('chat_disabled');
		
		/* hide private icon*/
		/*if (index != -1)
		{
			var div = domUsersList.children[index];
			var privateBtn = div.querySelector('.chat_pm_btn');
			if (privateBtn) privateBtn.style.visibility = 'hidden';
		}*/
		
		/* update ban icons and buttons*/
		var ban = objChat.isBanAllowed();
		var banDisplay = ban ? 'block' : 'none';
		domShowBannedBtn.style.display = banDisplay;
        domClearHistoryBtn.style.display = banDisplay;
        domAllowUnauthBtn.style.display = banDisplay;
		var users = domUsersList.querySelectorAll('.chat_user');
		for (var i = 0; i < users.length; i++)
		{
			var item = users[i];
			var id = item.getAttribute('data-id');
			updateBanBtn(item, id == user.id ? user : objChat.getUserByID(id), ban);
		}
		
		/* enable/disable poll*/
		objPoll.updateAdminMode();
	}

	/** Logout event handler. */
	function onLogout()
	{
        domChatContent.classList.remove('chat_logged_in');
		domMessageInput.setAttribute('contenteditable', 'false');
		domNickInput.setAttribute('contenteditable', 'false');
		/* disable buttons*/
		domSendBtn.classList.add('chat_disabled');
		domSmileBtn.classList.add('chat_disabled');
		domPollBtn.classList.add('chat_disabled');
		domSettingsBtn.classList.add('chat_disabled');
		/*if (bFirstLogin && !funcLoginHandler)
		{
			bFirstLogin = false;
			if (window.WSChat && !window.WSChat.checkSocialLogin())
				domChatContent.querySelector('.chat_login').classList.add('chat_showed');
		}*/
		domChatContent.querySelector('.chat_login').classList.add('chat_showed');
		objPoll.updateAdminMode();
        setDisplay(domPollBtn, false);
	}

	/** Set nick event handler. */
	function onSetNick(nick)
	{
		domNickInput.innerText = nick;
		
		if (isFirstLogin)
		{
			isFirstLogin = false;
			selectionNick();
		}
    }
    
    /**
     * On message sent
     * 
     * @param {bool} accepted 
     */
    function onMsgSent(accepted)
    {
        if (accepted)
        {
            domMessageInput.innerHTML = '';
            curMessage.isLocal = true;
            curMessage.time = Math.floor(Date.now() / 1000);
            onAddMessage(curMessage);
        }
        curMessage = null;
        domMessageInput.classList.remove('chat_disabled');
        domMessageInput.contentEditable = true;
        domMessageInput.focus();
        domSendBtn.classList.remove('chat_disabled');
    }

    function getTimeStr(date)
    {
        var hours = date.getHours();
        var minutes = date.getMinutes();
        return (hours > 9 ? hours : '0' + hours) + ':' + (minutes > 9 ? minutes : '0' + minutes)
    }

	/**
	 * Add message event handler. 
	 *
	 * @param {Object}  item	    The message object.
     * @param {bool}    isReverse   If true, then inserts the message in beginning
	 */
	function onAddMessage(item, isReverse)
	{
		if (item.private && domCurrentHistory.getAttribute("data-tab") != item.target_id) return false;
        
		var content = !item.private ? domHistory : domCurrentHistory;
        var div = document.createElement('div');
        var nTopPos = content.scrollHeight - content.offsetHeight/2;
		if (typeof(item.system) === 'undefined')
		{
            var msgID = 'data-uid="'+item.utype +'|'+ item.uid +'"';
            if (typeof item.id !== 'undefined') { msgID += ' data-id="'+ item.id +'"'; }

            let tmpMsg = null;
            let targetNode = null;
            if (item.isLocal)
            {
                div.setAttribute('data-local-id', item.local_id);
            }
            else if (!isReverse) 
            {
                for (var i = content.childNodes.length - 1; i >= 0; i--)
		        {
                    var node = content.childNodes[i];
                    var localID = node.getAttribute('data-local-id');
                    if (localID)
                    {
                        if (localID == item.local_id)
                            tmpMsg = node;
                    }
                    else
                    {
                        targetNode = node;
                        break;
                    }
                }
            }

            if (!targetNode) targetNode = isReverse ? content.firstChild : content.lastChild;

            var date = new Date(item.time*1000);
			var text = '<span class="chat_msg_text">'+ parseMessageText(item.text, !item.private) + '</span><span class="chat_msg_time">' + getTimeStr(date) + '</span>';
            var uid = item.uid + ':' + item.nick;
            var dateKey = date.getFullYear() +'-'+ (date.getMonth() + 1) +'-'+ date.getDate();
            var lastID = targetNode ? targetNode.getAttribute('data-uid') : '';
            var targetDate = targetNode ? targetNode.getAttribute('data-date') : '';
            var utype = item.private ? (item.target_id == item.uid ? 'priv' : 'my') : item.utype;
            if (bFullMessages)
            {
                div.innerHTML = '<div class="chat_msg chat_first chat_ut_' + utype + '" '+ msgID 
                    +'><div class="chat_usr chat_ut_' + utype + '">' + htmlSpecialChars(item.nick) + '</div>' + text + '</div>';
            }
            else
            {
                if (lastID != uid || targetDate != dateKey)
                {
                    div.innerHTML = '<div class="chat_usr chat_ut_' + utype + '">' + htmlSpecialChars(item.nick)
                        +'</div><div style="clear:both;"></div><div class="chat_msg chat_first chat_ut_' + utype + '" '+ msgID +'>' + text + '</div>';
                }
                else
                {
                    div.innerHTML = '<div class="chat_msg chat_ut_' + utype + '" '+ msgID +'>' + text + '</div>';
                }
            }

            div.setAttribute('data-uid', uid);
            div.setAttribute('data-date', dateKey);

            if (targetNode)
            {
                if (isReverse) /* insert into beginning*/
                {
                    content.insertBefore(div, targetNode);
                    if (lastID == uid && dateKey == targetDate)
                    {
                        /* move user nick node (and date node if exists) from prev*/
                        var insPos = div.firstChild;
                        while (targetNode.firstChild && !targetNode.firstChild.classList.contains('chat_msg')) /* move all before '.chat_msg'*/
                        {
                            div.insertBefore(targetNode.firstChild, insPos);
                        }
                        insPos.classList.add('chat_first');
                        targetNode.firstChild.classList.remove('chat_first');
                    }
                    else
                    {
                        if (dateKey == targetDate)
                        {
                            /* move date node from previous*/
                            var dateNode = targetNode.querySelector('.chat_msg_date');
                            if (dateNode)
                                div.insertBefore(dateNode, div.firstChild);
                        }
                        else
                        {
                            createDateNode(div, date);
                        }
                    }
                }
                else
                {
                    insertAfter(div, targetNode);
                    if (dateKey != targetDate)
                    {
                        createDateNode(div, date);
                    }
                }
            }
            else
            {
                content.appendChild(div);
                createDateNode(div, date);
            }

            if (tmpMsg)
            {
                deleteMessage(tmpMsg);
            }
		}
		else
		{
			content = domCurrentHistory;
			var text = item.text;
			if (domLanguage)
			{
				var domText = domLanguage.querySelector('[data-id='+text+']');
				if (domText) text = domText.innerHTML;
			}
			div.innerHTML = text;
            div.classList.add('system_msg');
            content.appendChild(div);
		}
		
        while (content.children.length > objChat.getMaxHistorySize()) 
        {
            deleteMessage(content.children[0]);
        }
		
        updateTopDate(content);

        /* scroll to bottom*/
		if (content == domCurrentHistory)
		{
            domCurrentHistory.customScrollBar.update();
            if (content.scrollTop + content.offsetHeight >= nTopPos)
            {
				content.scrollTop = content.scrollHeight;
			}
        }
		
		return true;
    }

    /**
     * Update top date
     */
    function updateTopDate(container)
    {
        var dates = container.querySelectorAll('.chat_msg_date');
        var node = null;
        var prevNode = null;
        var topDateNode = null;

        /* find top date node*/
        for (var i = 0; i < dates.length; i++)
        {
            var dateNode = dates[i];
            if (container.scrollTop > dateNode.parentNode.offsetTop - 2 * dateNode.offsetHeight)
            {
                prevNode = node;
                node = dateNode;
            }
            else
            {
                break;
            }
        }

        if (node)
        {
            if (container.scrollTop < node.parentNode.offsetTop)
            {
                /* hide fixed date*/
                topDateNode = container.parentNode.querySelector('.chat_top_fixed_date');
                if (topDateNode)
                {
                    if (prevNode) 
                    {
                        updateDateNode(topDateNode.firstChild, prevNode.firstChild);
                        var dh = 0.5 * node.offsetHeight;
                        topDateNode.style.opacity = Math.min(1, Math.max(0, (node.parentNode.offsetTop - container.scrollTop - dh) / (2 * node.offsetHeight - dh)));
                    }
                    else
                    {
                        topDateNode.classList.add('chat_hide');
                    }
                }
            }
            else
            {
                /* show fixed date*/
                topDateNode = container.parentNode.querySelector('.chat_top_fixed_date');
                if (topDateNode)
                {
                    topDateNode.classList.remove('chat_hide');
                    topDateNode.style.opacity = 1;
                    updateDateNode(topDateNode.firstChild, node.firstChild);
                }
                else /* create top node of date*/
                {
                    topDateNode = document.createElement('div');
                    topDateNode.classList.add('chat_msg_date');
                    topDateNode.classList.add('chat_top_fixed_date');
                    topDateNode.innerHTML = node.innerHTML;
                    container.parentNode.insertBefore(topDateNode, container);
                    var styles = getComputedStyle(container);
                    var contentWidth = container.clientWidth - parseFloat(styles.paddingLeft) - parseFloat(styles.paddingRight);
                    topDateNode.style.width = 'calc(100% - ' + (container.parentNode.offsetWidth - contentWidth + 2 * topDateNode.offsetLeft) + 'px)';
                }
            }
        }
    }

    function updateDateNode(target, src)
    {
        if (target.innerText != src.innerText)
        {
            target.innerText = src.innerText;
            var dateKey = src.getAttribute('data-date');
            if (dateKey)
                target.setAttribute('data-date', dateKey);
            else
                target.removeAttribute('data-date');
        }
    }
    
    /**
     * Createa a new date node
     * 
     * @param {*} target    Container of date node (message node)
     * @param {*} date      DateTime
     */
    function createDateNode(target, date)
    {
        var dateDiv = document.createElement('div');
        dateDiv.classList.add('chat_msg_date');
        dateDiv.innerHTML = getDateHTML(date);
        target.insertBefore(dateDiv, target.firstChild);
        return dateDiv;
    }
	
	/**
	 * Handler of private messages counter.
	 * 
	 * @param {Object} user		User of private dialog.
	 * @param {int} count		Number of unread messages from the specified user. 
	 * @param {int} total		Total number of unread private messages.
	 */
	function onPrivateMsgCount(user, count, total)
	{
		var domTab = getPrivateHistoryTab(user);
		if (domTab)
		{
			var domCounter = domTab.querySelector('.counter');
			domCounter.innerText = count;
			domCounter.style.display = count ? 'inline' : 'none';
			if (count)
			{
				playPrivateSound();
			}
			if (domPrivateTabs.scrollWidth > domPrivateTabs.clientWidth)
				domPrivateTabs.classList.add('scrolled');
			else
				domPrivateTabs.classList.remove('scrolled');
		}
	}
	
	/**
	 *	Replaces special chars. 
	 */
	function htmlSpecialChars(text)
	{
		var map = {
		    '&': '&amp;',
			'<': '&lt;',
			'>': '&gt;',
			'"': '&quot;',
			"'": '&#039;'
		  };
 		return text.replace(/[&<>"']/g, function(m) { return map[m]; });
	}
	
	/**
	 *	Searches smileys and generates HTML text. 
	 */
	function parseMessageText(text, isCommon)
	{
		/* htmlspecialchars*/
		text = htmlSpecialChars(text).replace(/%(\d+)%/g, function(str, id) {
				if (aSmileys && aSmileys[id])
				{
					var className = aSmileys[id]['class'];
					var img = aSmileys[id]['img'];
					if (!img) img = sBlankImgURL;
					return '<img src="'+ img + '" data-id="'+ id +'"'+ (className ? ' class="'+ className +'"' : '') +'>';
				}
				else
					return '';
			}).replace(/\\%\\/g, '%');
		if (isCommon && objChat.getNick())
		{
			let arr = text.split(htmlSpecialChars(objChat.getNick()));
			if (arr.length > 1)
			{	
				playPrivateSound();			
				if(domHistoryTabs.querySelector('.selected') != domHistoryTabs.querySelector('[data-tab="common"]')){
					updateCounterHistory(true);
				}

				return arr.join("<span class='chat_mynick'>" + objChat.getNick() + "</span>");
			}
		} 
		return text;
	}

	/** Clear history event handler. */
	function onClearHistory()
	{
		domHistory.innerHTML = "";
	}
	
	/** Deletion of a message event handler. */
	function onDeleteMessage(msgID, msgIndex)
	{
		var domMsg = domChatContent.querySelector('.chat_history [data-id="' +msgID + '"]');
		if (domMsg)
		{ 
			var bCurrent = isParent(domMsg.parentNode, domCurrentHistory); 
			deleteMessage(domMsg.parentNode);
			if (bCurrent) domCurrentHistory.customScrollBar.update();
		}
	}
	
	/**
	 * Deletes a message node. 
	 */
	function deleteMessage(domMessage)
	{
		var domNext = domMessage.nextSibling;
		if (domMessage.querySelector('.chat_usr'))
		{ 
			if (domNext)
			{
				if (!domNext.querySelector('.chat_usr'))
				{
                    /* copy nick node (and date node) to the next message of this user*/
					var domMsg = domNext.children[0];
                    while (domMessage.firstChild && !domMessage.firstChild.classList.contains('chat_msg')) /* move all children before '.chat_msg'*/
                    {
                        domNext.insertBefore(domMessage.firstChild, domMsg);
                    }
                    domMsg.classList.add('chat_first');
                }
                else if (!domNext.querySelector('.chat_msg_date'))
                {
                    var domDate = domMessage.querySelector('.chat_msg_date');
                    if (domDate)
                    {
                        domNext.insertBefore(domDate, domNext.firstChild);
                    }
                }
			}
		}
		domMessage.parentNode.removeChild(domMessage);
    }

    var spareUserItems = [];
    const numCachedItems = 10;
    
    /**
     * Handler of users scroll
     */
    function onUsersScroll() 
    {
        var items = domUsersList.children;
        var count = objChat.usersCount();
        
        var h = items.length ? items[0].offsetHeight : 1;
        var top = Math.floor(domUsers.scrollTop / h);
        var btm = top + Math.floor(domUsers.clientHeight / h) + numCachedItems;
        top -= numCachedItems;
        if (top < 0) top = 0;
        if (btm >= count) btm = count - 1;
        var ind = 0;
        for (let i = top; i <= btm; i++)
        {
            let usr = objChat.getUserByInd(i);
            let found = false;
            for (let j = ind; j < items.length; j++)
            {
                if (items[j].getAttribute('data-id') == usr.id)
                {
                    if (parseInt(items[j].style.top, 10) !=  i * h)
                    {
                        items[j].style.top = (i * h) + 'px';
                    }
                    if (j != ind)
                    {
                        domUsersList.insertBefore(items[j], items[ind]);
                    }
                    ind++;
                    found = true;
                    break;
                }
            }
            if (!found)
            {
                let el = getUserItem(usr, i, h);
                domUsersList.insertBefore(el, items[ind]);
                ind++;
            }
        }

        while (ind < items.length)
        {
            releaseUserItem(ind);
        }
    }

    /**
     * Relese dom element of user
     * @param {*} pos   Position in users display list
     */
    function releaseUserItem(pos)
    {
        spareUserItems.push(domUsersList.children[pos]);
        domUsersList.removeChild(domUsersList.children[pos]);
    }

    /**
     * Acquire dom element for user item
     * @param {*} user 
     * @param {*} pos 
     * @param {*} itemHeight 
     */
    function getUserItem(user, pos, itemHeight)
    {
        var div;
        var nick;
        var pmBtn;
        if (spareUserItems.length)
        {
            div = spareUserItems.pop();
            nick = div.querySelector('.chat_nick');
            pmBtn = div.querySelector('.chat_pm_btn');
        }
        else
        {
            div = document.createElement('div');
            var nickWrapper = document.createElement('div');
            nick = document.createElement('div');
            nick.className = 'chat_nick';
            nickWrapper.className = 'chat_nick_wrapper';
            nickWrapper.appendChild(nick);
            div.appendChild(nickWrapper);
            /* personal msg btn*/
            var pmBtn = document.createElement('div');
            pmBtn.className = 'chat_pm_btn';
            div.appendChild(pmBtn);
            /* ban btn*/
            var banBtn = document.createElement('div');
            banBtn.className = 'chat_ban_btn';
		    div.appendChild(banBtn);
        }
        nick.innerText = user.nick;
        pmBtn.style.visibility = user.current ? 'hidden' : 'visible';
        div.className = 'chat_user chat_ut_' + user.type;
        div.style.top = (pos * itemHeight) + 'px';
        div.setAttribute('data-id', user.id);
        updateBanBtn(div, user, objChat.isBanAllowed());
        return div;
    }

	/** Add user event handler. */
	function onAddUser(user, pos, update)
	{
        var div = null;
        /* detect visibility*/
        var updateVisible = true;
        let h = domUsersList.children.length ? domUsersList.children[0].offsetHeight : 1;
        let top = Math.floor(domUsers.scrollTop / h);
        var btm = top + Math.floor(domUsers.clientHeight / h) + numCachedItems;
        top -= numCachedItems;
        if (top < 0) top = 0;
        if (pos < top || pos > btm)
        {
            if (pos > top) updateVisible = false;
        }
        else
        {
            div = getUserItem(user, pos, h);
            var p = pos - top;
            if (p < domUsersList.children.length)
            {
                domUsersList.insertBefore(div, domUsersList.children[p]);
                while (++p < domUsersList.children.length)
                {
                    if (p > btm)
                    {
                        releaseUserItem(p);
                    }
                    else
                    {
                        domUsersList.children[p].style.top = ((p + top) * h) + 'px';
                    }
                }
            }
            else
            {
                domUsersList.appendChild(div);
            }
            updateVisible = false;
        }

        if (update)
        {
            domUsersList.style.height = (h * objChat.usersCount()) +'px';
            domUsers.customScrollBar.update();
            if (updateVisible) onUsersScroll();
        }
	}

	/** Remove user event handler. */
	function onRemoveUser(user, pos)
	{
        let h = domUsersList.children.length ? domUsersList.children[0].offsetHeight : 1;
		onUsersScroll();
        domUsersList.style.height = (h * objChat.usersCount()) +'px';
        domUsers.customScrollBar.update();
	}

	/** Update user event handler. */
	function onUpdateUser(user, oldPos, newPos)
	{
        onUsersScroll();
        var items = domUsersList.children;
        for (let i = 0; i < items.length; i++)
        {
            var item = items[i];
            if (item.getAttribute('data-id') == user.id)
            {
                var nick = item.querySelector('.chat_nick');
                nick.innerText = user.nick;
			    item.className = 'chat_user chat_ut_' + user.type;
                updateBanBtn(item, user, objChat.isBanAllowed());
                break;
            }
        }
		/*if (oldPos < domUsersList.children.length)
		{
			var div = domUsersList.children[oldPos];
			div.children[0].children[0].innerText = user.nick;
			div.className = 'chat_user chat_ut_' + user.type;
			var ban = objChat.isBanAllowed(); 
			updateBanBtn(div, user, ban);
			if (!ban) div.style.display = user.banned ? "none" : "block";
			
			if (oldPos != newPos)
			{
				domUsersList.removeChild(domUsersList.children[oldPos]);
				if (newPos < domUsersList.children.length)
					domUsersList.insertBefore(div, domUsersList.children[newPos]);
				else
					domUsersList.appendChild(div);
			}
		}
		else
		{
			onAddUser(user, newPos);
		}*/
	}
	
	function onHoverUser()
	{
		var cUsersCont = document.querySelector(".chat_users");
		cUsersCont.addEventListener('mouseover', function(e){
			if (e.target.className == "chat_nick")
			{
				var th = e.target.closest(".chat_user");
				var domNick = th.querySelector(".chat_nick");
                if (domNick.classList.contains("animate") || th.offsetWidth >= domNick.offsetWidth ) return;
                var domWrapper = th.querySelector(".chat_nick_wrapper");
				var nAnimTime = domNick.offsetWidth/200;
				domNick.classList.add("animate");
				domNick.style.transition = "left " + nAnimTime + "s";
				domNick.style.left = "-" + (domNick.offsetWidth  - domWrapper.offsetWidth + 5) +'px';
                /*setTimeout(function(){domNick.classList.remove("animate"); domNick.style.left = 0;}, nAnimTime*1000);*/
                var onOut = function() {
                    domNick.removeEventListener('mouseout', onOut);
                    domNick.classList.remove('animate');
                    domNick.style.left = 0;
                };
                domNick.addEventListener('mouseout', onOut);
			}
		});
	}
	
	onHoverUser();
	
	/**
	 * Updates ban button in user item. 
	 */
	function updateBanBtn(div, user, banAllowed)
	{
		var btn = div.querySelector('.chat_ban_btn');
        btn.style.visibility = user.current ? 'hidden' : 'visible';
        
		if (banAllowed && objChat.isUserBan(user))
		{
			btn.classList.remove('chat_mute_btn');
		}
		else
		{	
            btn.classList.add('chat_mute_btn');
			if (user.muted)
			{
				 div.classList.add('chat_muted');
				 closeHistoryPage(user.id);
			} 
            else
            {
                div.classList.remove('chat_muted');
            }
		}
		
		if (user.banned)
		{
			div.classList.add('chat_banned');
		}
		else
		{
			div.classList.remove('chat_banned');
		}
	}
	
	/**
	 *	Returns text from input element. 
	 */
	function getInputString(domInput)
	{
		var message = '';
		for (var i = 0; i < domInput.childNodes.length; i++)
		{
			var node = domInput.childNodes[i];
			if (node.nodeType == 1) /* element node*/
			{
				var id = node.getAttribute('data-id');
				if (id)
				{
					message += "%"+ id +"%";
					continue;
				}
			}
			message += node.textContent.replace(/%/g, "\\%\\");
		}
		return message.trim();
	}

	/**
	 * Sends a new message.
	 */
	function SendMessage()
	{
        if (curMessage) return;
		var message = getInputString(domMessageInput);	
		if (message != '')
		{	
            var targetID = domCurrentHistory.getAttribute("data-tab");
            if (targetID == 'common') { targetID = null; }
            var localId = objChat.sendMessage(message, targetID);
            if (localId)
            {
                var user = objChat.getUser();
                curMessage = {
                    local_id: localId,
                    text: message,
                    utype: user.type,
                    uid: user.id,
                    nick: user.nick,
                    target_id: targetID,
                    private: targetID !== null
                };
                domMessageInput.classList.add('chat_disabled');
                domMessageInput.contentEditable = false;
                domSendBtn.classList.add('chat_disabled');
            }
		}
	}
	
	/**
	 *	Returns current number of smileys in the message input. 
	 */
	function calcNumSmileys()
	{
		var nSmileys = 0;
		for (var i = 0; i < domMessageInput.childNodes.length; i++)
		{
			var node = domMessageInput.childNodes[i];
			if (node.nodeType == 1 && node.getAttribute('data-id')) /* smile node*/
				nSmileys++;
		}
		return nSmileys;
	}
	
	/**
	 *	Limits smileys input. 
	 */
	function onMessageInputChange()
	{
		var nSmileys = 0;
		for (var i = 0; i < domMessageInput.childNodes.length; i++)
		{
			var node = domMessageInput.childNodes[i];
			if (node.nodeType == 1 && node.getAttribute('data-id')) /* smile node*/
			{
				nSmileys++;
				if (nSmileys > nMaxSmileysInMsg) domMessageInput.removeChild(node);
			}
		}
	}

	/**
	 *	Changes the user nickname.
	 */
	function SendNick()
	{
		objChat.setNick(domNickInput.innerText);
	}
	
	/**
	 * Handler of smile click.
	 * @param {Object} domSmile
	 * @param {string} id
	 */
	function onClickSmile(domSmile, id)
	{
		if (calcNumSmileys() >= nMaxSmileysInMsg) return;
		
		var node = document.createElement('img');
        node.className = domSmile.className;
        node.setAttribute('data-id', id);
        node.src = sBlankImgURL;
		insertInputNode(node, '%' + id + '%');
		domSmileysDlg.classList.remove('chat_showed');
	}
		
	/**
	 * Sends HTTP GET request. 
 	 * @param {Object} url
 	 * @param {Function} callback
	 */
	/*function httpGet(url, callback)
	{
	    var xmlHttp = new XMLHttpRequest();
	    xmlHttp.onreadystatechange = function() { 
	        if (xmlHttp.readyState == 4 && xmlHttp.status == 200)
	            callback(xmlHttp.responseText);
	    };
	    xmlHttp.open("GET", url, true);
	    xmlHttp.send(null);
	}*/
	
	/**
	 * Handler of click event on users list.
	 * @param {Object} event
	 */
	function onUsersClick(event)
	{
		if (event.target.classList.contains('chat_mute_btn'))
		{
			showMuteDlg(event.target, event.target.parentNode.getAttribute('data-id'));
        }
        else if (event.target.classList.contains('chat_ban_btn')) /* ban icon*/
		{
			showBanDlg(event.target, event.target.parentNode.getAttribute('data-id'));
		}
		else if (event.target.classList.contains('chat_pm_btn')) /* private messgae icon*/
		{
			var userID = event.target.parentNode.getAttribute('data-id');
			if (getPrivateHistoryTab(objChat.getUserByID(userID)))
				showHistoryPage(userID);
		}
		else if (event.target.classList.contains('chat_nick')) /* nick click*/
		{
			var text = event.target.textContent;
			insertInputNode(document.createTextNode(text), text);
		}
	}
	
	/**
	 * Handler of ban button click.  
	 * @param {Object} event
	 */
	function onBanUserClick(event)
	{
		var action = event.target.getAttribute('data-action');
		var userID = event.target.parentNode.getAttribute('data-id');
		switch (action)
		{
			case 'mute':
			case 'unmute':
				objChat.setUserMute(userID, action == 'mute');
				break;
			default:
				objChat.setUserBan(userID, event.target.getAttribute('data-action'));
		}
		event.target.parentNode.classList.remove('chat_showed');
    }
    
    /**
     * Returns user nick by message node
     * @param {*} domMessage 
     */
    function getUserNickFromMessage(domMessage)
    {
        domMessage = domMessage.parentNode;
        while (domMessage)
        {
            var domUser = domMessage.querySelector('.chat_usr');
            if (domUser)
            {
                return domUser.textContent;
            }
            domMessage = domMessage.previousSibling;
        }
        return '';
    }
	
	/**
	 * Show ban dialog. 
	 * @param {Object} target
	 * @param {string} userID
	 */
	function showBanDlg(target, userID, domMessage)
	{
		var user = objChat.getUserByID(userID);
		var nick = '';
		var banned = false;
		if (user)
		{
			nick = user.nick;
			banned = user.banned;
		}
		else if (domMessage)
		{
			/* get user from message*/
			nick = getUserNickFromMessage(domMessage);
		}
		domBanDlg.querySelector(".chat_title > span").innerText = nick;
		
		/* calculate dialog position*/
		placeBanDlg(target, domBanDlg);
		
		domBanDlg.setAttribute('data-id', userID);
		
		/* show/hide ban/unban buttons*/
		var banButtons = domBanDlg.querySelectorAll('.chat_push_btn');
		for (var i = 0; i < banButtons.length; i++)
		{
			var action = banButtons[i].getAttribute('data-action');
			banButtons[i].style.display = action == "unban" ? (banned ? "block" : "none") : (banned  ? "none" : "block");
		}
	}
	
	/**
	 *	Calculates dialog position by target element. 
	 */
	function placeBanDlg(target, dlg)
	{
		var trect = target.getBoundingClientRect();
		var prect = dlg.parentNode.getBoundingClientRect();
		var left = trect.left - prect.left;
		var top = trect.top - prect.top;
		var margin = 20; 
		dlg.classList.add('chat_showed');
		if (left + dlg.offsetWidth + margin > dlg.parentNode.offsetWidth)
			left = dlg.parentNode.offsetWidth - dlg.offsetWidth - margin;
		if (top + dlg.offsetHeight > dlg.parentNode.offsetHeight)
			top = dlg.parentNode.offsetHeight - dlg.offsetHeight;
		dlg.style.left = left + "px";
		dlg.style.top = top + "px";
	}
	
	/**
	 * Show ban dialog. 
	 * @param {Object} target
	 * @param {string} userID
	 */
	function showMuteDlg(target, userID)
	{
		var user = objChat.getUserByID(userID);
		if (!user) return;
		domMuteDlg.querySelector(".chat_title > span").innerText = user.nick;
		placeBanDlg(target, domMuteDlg);
		domMuteDlg.setAttribute('data-id', userID);
		/* show/hide buttons*/
		var buttons = domMuteDlg.querySelectorAll('.chat_push_btn');
		for (var i = 0; i < buttons.length; i++)
		{
			var action = buttons[i].getAttribute('data-action');
			buttons[i].style.display = (action == 'mute' && !user.muted) || (action == 'unmute' && user.muted) ? 'block' : 'none';
		}
	}
		
	
	/**
	 * Handler of ban event for current user.
 	 * @param {bool} banned
	 */
	function onBan(banned)
	{
		domChatContent.querySelector('.chat_ban_banner').style.display = banned ? "block" : "none";
	}
	
	/**
	 * Handler of click event on login buttons.
	 *  
 	 * @param {Object} event
	 */
	function onLoginClick(event)
	{
		/*var type = event.target.getAttribute("data-type"); 
		if (type && window.WSChat)
			window.WSChat.openAuthWindow(type);*/
		login_Main();
	}
	
	/**
	 * Handler of click event on 'show banned' settings option.
	 *  
 	 * @param {Object} event
	 */
	function onShowBannedClick(event)
	{
		domShowBannedBtn.classList.toggle('chat_checked');
		var bVisible = domShowBannedBtn.classList.contains('chat_checked');
		objChat.setBannedVisible(bVisible);
		/*var users = domUsersList.querySelectorAll(':scope > div.chat_banned');
		for (var i = 0; i < users.length; i++)
				users[i].style.display = bVisible ? 'block' : 'none';*/
	}
	
	/**
	 * Handler of click event on 'show moderator' settings option.
	 *  
 	 * @param {Object} event
	 */
	function onShowModeratorClick(event)
	{
		domShowModeratorBtn.classList.toggle('chat_checked');
		var bVisible = domShowModeratorBtn.classList.contains('chat_checked');
		objChat.setModeratorVisible(bVisible);
	}
	
	/**
	 * Updates a left position of the poll dialog by the poll button position. 
	 */
	function updatePollDlgPosition()
	{
		var left = domPollBtn.offsetLeft;
		var style = getComputedStyle(document.documentElement);
		var shift = parseInt(style.getPropertyValue('--vchat-poll-dlg-shift'));
		if (!isNaN(shift)) left += shift;
		domPollDlg.style.left = left + "px";
	}
	
	/**
	 * Chat poll GUI.
	 */
	function ChatPollGUI()
	{
		/* assign events*/
		objChat.onPoll			= onPoll;
		objChat.onPollResults	= onPollResults;
		objChat.onPollParams	= onPollParams;
		
		/* private variables*/
		var domQuestionInput = domPollDlg.querySelector('.chat_poll_question');
		var domStartBtn = domPollDlg.querySelector('.chat_poll_start_btn');
		var domShowResultsCheck = domPollDlg.querySelector('.chat_poll_control .chat_poll_show_results');
		var aVariants = [];
		domStartBtn.addEventListener('click', onStartClick);
		domQuestionInput.addEventListener('focus', onInputFocus);
		domQuestionInput.addEventListener("keypress", function(event){if (event.keyCode == 10 || event.keyCode == 13) event.preventDefault();});
		if (objChat.isPollShowResults()) domShowResultsCheck.classList.add('chat_checked');
        domShowResultsCheck.addEventListener('click', onShowResultClick);
        
        function updateScrollBar()
        {
            domPollDlg.querySelector('.chat_poll_questions').customScrollBar.update();
        }
				
		/**
		 * Add event handlers for answer input. 
		 */
		function addVariantEventHandlers(domInput)
		{
			if ("MutationObserver" in window)
			{
				var objInputObserver = new MutationObserver(onVariantInputChange);
				objInputObserver.observe(domInput, {attributes: false, childList: true, characterData: true, subtree: true});
			}
			else
				domInput.addEventListener('DOMSubtreeModified', onVariantInputChange);
			domInput.addEventListener('focus', onInputFocus);
		}
		
		/**
		 * Handler of variant focus event. 
		 */
		function onInputFocus(event)
		{
			event.target.classList.remove('chat_alert');
		}
		
		/**
		 * Handler of click on an answer variant. 
		 */
		function onVariantClick(event)
		{
			if (objChat.getPollAnswerIndex() != -1 ||
				objChat.getPollStatus() != 'started' ||
				objChat.isCanEditPoll())
				return;
			
			var target = event.target;
			while (target && !target.classList.contains('chat_poll_input'))
				target = target.parentNode;
			
			if (target)
			{
				var index = parseInt(target.getAttribute('data-index'));
				if (objChat.answerPoll(index))
				{
					domPollDlg.classList.remove('chat_poll_answer');
					if (objChat.isPollShowResults()) domPollDlg.classList.add('chat_poll_results');
					aVariants[index].parentNode.classList.add('chat_poll_selected');
				}
            }
            updateScrollBar();
		}
		
		/**
		 * Handler of click on the "show results" checkbox. 
		 */
		function onShowResultClick(event)
		{
			domShowResultsCheck.classList.toggle('chat_checked');
			objChat.setPollShowResults(domShowResultsCheck.classList.contains('chat_checked'));
		}
		
		/**
		 * Handler of change event of poll answer input. 
		 */
		function onVariantInputChange(event)
		{
			if (!objChat.isCanEditPoll() || objChat.getPollStatus() != '') return;
			
			var domLastEmpty = null;
			var nEmpty = 0;
			for (var i = 0; i < aVariants.length; i++)
				if (getInputString(aVariants[i]) == '')
				{
					domLastEmpty = aVariants[i];
					nEmpty++;
				}
			
			if (domLastEmpty === null)
			{
				if (aVariants.length < objChat.getMaxPollAnswers())
				{
					/* create new variant input*/
					createVariantItem(objChat.getPollStatus());
				}
			}
			else if (aVariants.length > 3 && nEmpty > 1)
			{
				aVariants.splice(aVariants.indexOf(domLastEmpty), 1);
				domLastEmpty.parentNode.parentNode.removeChild(domLastEmpty.parentNode);
            }
            updateScrollBar();
		}
		
		/**
		 * Creates an answer item. 
		 */
		function createVariantItem(pollStatus, answerText)
		{
			var div = document.createElement('div');
			var domTemplate = domPollDlg.querySelector('.chat_poll_variant_template');
			div.innerHTML = domTemplate.innerHTML;
			domTemplate.parentNode.appendChild(div);
			div.classList.add('chat_poll_variant');
			var input = div.querySelector('.chat_poll_input');
			input.addEventListener("keypress", function(event){if (event.keyCode == 10 || event.keyCode == 13) event.preventDefault();});
			if (objChat.isCanEditPoll() && pollStatus == '')
			{ 
				addVariantEventHandlers(input);
				input.setAttribute('contenteditable', 'true');
			}
			if (answerText)
			{
				input.innerHTML = parseMessageText(answerText);
				input.setAttribute('data-index', aVariants.length);
				input.addEventListener('click', onVariantClick);
			}
            aVariants.push(input);
            updateScrollBar();
		}
		
		/**
		 * Sets editing mode. 
		 */
		this.updateAdminMode = function()
		{
			var bEdit = objChat.isCanEditPoll(); 
			var sStatus = objChat.getPollStatus(); 
			if (bEdit) domPollDlg.classList.add('chat_poll_admin'); else domPollDlg.classList.remove('chat_poll_admin');
			if (bEdit && sStatus == '') domPollDlg.classList.add('chat_poll_edit'); else domPollDlg.classList.remove('chat_poll_edit');
			if (!bEdit && sStatus == 'started') domPollDlg.classList.add('chat_poll_answer'); else domPollDlg.classList.remove('chat_poll_answer');
			setDisplay(domPollBtn, objChat.isCanEditPoll() ? true : objChat.isPollStarted());
            if (bEdit) domPollDlg.classList.add('chat_poll_results'); else onPollParams('show_results', objChat.isPollShowResults());
            updateScrollBar();
		};
		
		/**
		 * Handler of poll actions (start, stop, clear). 
		 */
		function onPoll(objPoll)
		{
			var i;
			var pollStatus = objPoll ? objPoll.status : '';
			if (objPoll === false) domPollDlg.classList.add('chat_poll_disabled'); else domPollDlg.classList.remove('chat_poll_disabled');
			
			/* clear answers items*/
			aVariants = [];
			var variants = domPollDlg.querySelectorAll('.chat_poll_variant');
			for (var i = 0; i < variants.length; i++)
				variants[i].parentNode.removeChild(variants[i]);
			domQuestionInput.setAttribute('contenteditable', 'false');
			domPollDlg.classList.remove('chat_poll_edit');
			domPollDlg.classList.remove('chat_poll_answer');
			domPollDlg.classList.remove('chat_poll_results');
						
			setDisplay(domPollBtn, objChat.isCanEditPoll() ? true : objChat.isPollStarted());
			var answerIndex = objChat.getPollAnswerIndex();
			/* fill answers*/
			switch (pollStatus)
			{
				case 'started':
					if (!objChat.isCanEditPoll())
					{
						if (answerIndex == -1)
						{
							/*updatePollDlgPosition();*/
							domPollDlg.classList.add('chat_showed');
							domPollDlg.classList.add('chat_poll_answer');
						}
						else if (objChat.isPollShowResults())
						{
							domPollDlg.classList.add('chat_poll_results');
						}
					}
					else
					{
						domPollDlg.classList.add('chat_poll_results');
					}
				break;
				
				case 'stopped':
					if (objChat.isCanEditPoll() || objChat.isPollShowResults())
						domPollDlg.classList.add('chat_poll_results');
				break;
				
				default: 
					/* new poll, create variant inputs*/
					pollStatus = '';
					if (objChat.isCanEditPoll())
					{ 
						domPollDlg.classList.add('chat_poll_edit');
						domQuestionInput.setAttribute('contenteditable', 'true');
					}
					else
					{
						domPollDlg.classList.remove('chat_showed');
					}
					for (i = 0; i < 3; i++)
						createVariantItem(pollStatus);
			}
			
			/* show variants */	
			if (pollStatus != '')
			{
				domQuestionInput.innerHTML = parseMessageText(objPoll.question);
				for (i = 0; i < objPoll.answers.length; i++)
				{
					createVariantItem(pollStatus, objPoll.answers[i]);
					if (i == answerIndex)
						aVariants[i].parentNode.classList.add('chat_poll_selected');
				}
			}
			
			/* update start button*/
			domStartBtn.classList.remove('chat_poll_start');
			domStartBtn.classList.remove('chat_poll_stop');
			domStartBtn.classList.remove('chat_poll_new');
			domStartBtn.classList.remove('chat_disabled');
			switch (pollStatus)
			{
				case 'started':		domStartBtn.classList.add('chat_poll_stop'); break;
				case 'stopped':		domStartBtn.classList.add('chat_poll_new'); break;
				default: 			domStartBtn.classList.add('chat_poll_start'); break;
			}
			
			/* update results*/
			onPollResults(objPoll ? objPoll.results : null);
		}
		
		/**
		 * Hanler of change of poll results. 
		 */
		function onPollResults(aPollResults)
		{
			var i;
			if (!aPollResults)
			{
				aPollResults = [];
				for (i = 0; i < aVariants.length; i++)
					aPollResults.push(0);
			}	
			
			var len = aPollResults.length;
			var nSum = 0;
			
			for (i = 0; i < len; i++)
				nSum += aPollResults[i];
				
			for (i = 0; i < len; i++)
			{
				var variant = aVariants[i].parentNode;
				var percent = nSum ? Math.round(100*aPollResults[i]/nSum) : 0;
				variant.querySelector('.chat_poll_result_fill').style.width = percent + '%';
				variant.querySelector('.chat_poll_result_text').innerText = aPollResults[i] + '(' + percent + '%)';
			}
		}
		
		/**
		 * Handler of change poll parameter.
		 * @param name	The parameter name.
		 * @param value The parameter value. 
		 */
		function onPollParams(name, value)
		{
			switch(name)
			{
				case 'show_results':
					if (value)
						domShowResultsCheck.classList.add('chat_checked');
					else
						domShowResultsCheck.classList.remove('chat_checked');
					if (!objChat.isCanEditPoll())
					{
						if ((objChat.getPollStatus() == 'started' && objChat.getPollAnswerIndex() != -1)
							|| objChat.getPollStatus() == 'stopped')
						{
							if (value)
								domPollDlg.classList.add('chat_poll_results');
							else
								domPollDlg.classList.remove('chat_poll_results');
						}
                    }
                    updateScrollBar();
				break;
			}
        }
		
		/**
		 * Handler of start button click. 
		 */
		function onStartClick(event)
		{
			switch (objChat.getPollStatus())
			{
				case 'started':
				case 'stopped':
					objChat.stopPoll();
				break;
				
				default:
					/* start new poll*/
					var questionText = getInputString(domQuestionInput);
					if (questionText == '')
					{
						domQuestionInput.classList.add('chat_alert');
						return;
					}
					var aVariants = domPollDlg.querySelectorAll('.chat_poll_variant>.chat_poll_input');
					var aAnswers = [];
					var domFirstEmpty = null;
					for (var i = 0; i < aVariants.length; i++)
					{
						var text = getInputString(aVariants[i]);
						if (text != '')
							aAnswers.push(text);
						else if (domFirstEmpty == null)
							domFirstEmpty = aVariants[i];
					}
					if (aAnswers.length < 2)
					{
						if (domFirstEmpty) domFirstEmpty.classList.add('chat_alert');
						return;
					}
					objChat.startPoll(questionText, aAnswers);
				break;
			}
			domStartBtn.classList.add('chat_disabled');
		}
	}
	
	function playPrivateSound()
	{

		if (domHearMsgSoundBtn.classList.contains('chat_checked') && (objMsgBlipSound || sMsgBlipSoundURL)){
			if (!objMsgBlipSound)
				objMsgBlipSound = new Audio(sMsgBlipSoundURL);
			objMsgBlipSound.play();
		}
	}

	function updateCounterHistory(newMessage)
	{

		let check = domHistoryTabs.querySelector('.selected') == domHistoryTabs.querySelector('[data-tab="common"]');
		check = newMessage ? check : !check;
		let counterHistory = domHistoryTabs.querySelector('.newcounter');
		countNewMassege = check ? 0 : (countNewMassege + 1);
		counterHistory.style.display = countNewMassege ? 'inline' : 'none';
		counterHistory.innerText = countNewMassege ;		
	}

	function selectionNick()
	{
		let messages = domHistory.querySelectorAll('.chat_msg_text');
		for (var i = 0; i < messages.length; i++)
		{
			let el = messages[i];
			let arr = el.innerText.split(htmlSpecialChars(objChat.getNick()));
			if (arr.length > 1)
			{
				el.innerHTML = arr.join("<span class='chat_mynick'>" + objChat.getNick() + "</span>");
			}
		}
	}
	/**
	 * Returns current nick.
	 */
	this.getNick = function()
	{
		return objChat.getNick();
	};

	this.disconnect = function()
	{
		objChat.disconnect();
	};
}
