﻿/**
 *	Live connection class.
 */
function LiveConnection(onConnect, onMessage, onData, timeout) {
    /* capture context*/
    var _connection = this;

	/*
	 *	PRIVATE
	 */

    var objSocket = null;
    var sURL = "";
    var nReconnectTimeoutID = -1;
    var nReconnectTimeout = timeout || 7000;
    var bConnected = false;

	/**
	 *	Stops the socket reconnection.
	 */
    function stopReconnection() {
        if (nReconnectTimeoutID != -1) {
            clearTimeout(nReconnectTimeoutID);
            nReconnectTimeoutID = -1;
        }
    }

	/**
	 *	Starts the socket reconnection with timeout.
	 */
    function reconnectSocket() {
        nReconnectTimeoutID = setTimeout(function () {
            if (objSocket) {
                var socket = objSocket;
                deleteSocket();
                socket.close();
            }
            _connection.connect(sURL);
        }, nReconnectTimeout);
    }

	/**
	 *	Delete websocket object.
	 */
    function deleteSocket() {
        if (objSocket) {
            objSocket.onopen = null;
            objSocket.onclose = null;
            objSocket.onmessage = null;
            objSocket.onerror = null;
            objSocket = null;
        }
    }

	/*
	 *	PUBLIC
	 */

	/**
	 *	Connects to the server.
	 */
    this.connect = function (url, binaryType) {
        sURL = url;
        objSocket = new WebSocket(url);
        objSocket.binaryType = binaryType || "arraybuffer";
        stopReconnection();
        /*if( navigator.userAgent.toLowerCase().indexOf('firefox') < 0 ) reconnectSocket();*/

        /* Handler of open event. */
        objSocket.onopen = function () {
            stopReconnection();
            bConnected = true;
            if (onConnect) onConnect(true);
        };

        /* Handler of close event. */
        objSocket.onclose = function (event) {
            bConnected = false;
            if (onConnect) onConnect(false);
            deleteSocket();
            reconnectSocket();
        };

        /* Handler of message event. */
        objSocket.onmessage = function (event) {
            if (typeof (event.data) == 'string') {
                /* text data (commands)*/
                var msg;
                try { msg = JSON.parse(event.data); }
                catch (error) {
                    console.log('wrong message: ' + event.data);
                    return;
                }

                if (msg.cmd === "redirect") {
                    console.log('Redirect connection: ' + msg.host);
                    var a = document.createElement("a");
                    a.href = sURL;
                    a.hostname = msg.host;
                    _connection.disconnect();
                    _connection.connect(a.href);
                    return;
                }
                if (onMessage) onMessage(msg);
            }
            else /* binary data*/
            {
                if (onData) onData(event.data);
            }
        };

        /* Handler of error event. */
        objSocket.onerror = function (error) {
            if (bConnected) {
                bConnected = false;
                if (onConnect) onConnect(false);
            }
            console.log("Websocket error: " + error.message);
            deleteSocket();
            reconnectSocket();
        };
    };

	/**
	 *	Current URL accessor.  
	 */
    this.getURL = function () { return sURL; };

    this.send = function (data) {
        if (bConnected) { objSocket.send(data); return true; }
        return false;
    };

    this.isConnected = function () { return bConnected; };

    this.disconnect = function (reconnect) {
        var socket = objSocket;
        if (!reconnect) {
            deleteSocket();
            stopReconnection();
        }
        if (socket && bConnected) {
            socket.close();
            if (!reconnect && onConnect) onConnect(false);
        }
        bConnected = false;
    };
}