/**
 *	Events synchronization class.
 */
function LiveSync(params)
{
    var nReconnectWSTimeout     = 5000;
    var objConnection = new LiveConnection(onWebscoketConnect, onWebsocketMessage, null, nReconnectWSTimeout);
   
    /**
     *  Subscribe to event
     */
    this.subscribe = function(name, id, sign) {
        if (objConnection.isConnected()) {
            objConnection.send(JSON.stringify({
                cmd: 'subscribe', 
                event: name,
                id: id,
                sign: sign
            }));
        }
    };

    /**
     *  Subscribe to event
     */
    this.unsubscribe = function(name, id) {
        if (objConnection.isConnected()) {
            objConnection.send(JSON.stringify({
                cmd: 'unsubscribe', 
                event: name,
                id: id
            }));
        }
    };

    /**
     *  Dispatch notification
     */
    this.notify = function(name, data, id, sign) {
        if (objConnection.isConnected()) {
            objConnection.send(JSON.stringify({
                cmd: 'notify', 
                event: name,
                id: id,
                sign: sign,
                data: data
            }));
        }
    };

    /**
     *  Dispatch notification for multiple IDs
     */
    this.notifyMulti = function(name, data, ids) {
        if (objConnection.isConnected()) {
            objConnection.send(JSON.stringify({
                cmd: 'notify', 
                event: name,
                ids: ids,
                data: data
            }));
        }
    };

    if (!params.url) {
        console.log('LiveSync url not specified');
        return;
    }

    objConnection.connect(params.url);

    this.disconnect = function()
	{
		objConnection.disconnect();
	};

    /**
	 *	Handler of connection/disconnection.
	 *	@param connected	Boolean flag.
	 */
	function onWebscoketConnect(connected)
	{
		if (connected) objConnection.send(JSON.stringify({cmd: 'login', login: params.login, password: params.password, role: params.role}));
	}
	
	/**
	 *	Processes server commands.
	 *	@param msg	A message object, received from the server.
	 */
	function onWebsocketMessage(msg)
	{
		switch (msg.cmd)
		{
			case 'login':
				if (params.onReady) params.onReady(msg);
				break;
            
            case 'notification':
                if (params.onNotification) params.onNotification(msg);
                break;

            case 'error':
                if (params.onError)
                    params.onError(msg);
                else
                    console.error('sync error: ', msg.code, msg.name);
                break;
		}
    }

}