var chatGUI = null;

/**##################################################################################################################################
 *	On ready handler. 
 * ##################################################################################################################################
 */
function	chat_Ready( cChatURL, aSmileys )
{
	/* connection info*/
	var domChatContainer	= document.getElementById('chat_wrapper');
	var sChannelID			= document.getElementById('chat_channel_id').innerHTML;
	var sURL				= cChatURL;
	var sUserID				= document.getElementById('chat_user_id').innerHTML;
	var domUSettinsURL		= null;
	var sUSettingsURL		= domUSettinsURL ? domUSettinsURL.innerHTML : '';
	/* initialize chat interface*/
	chatGUI = new ChatGUI(domChatContainer, sURL, sUSettingsURL, (sUserID != '' ? loginHandler : null),
			aSmileys,
			"img/chat/blank.gif",
			"msg_blip.mp3",
			new MobileChatExtension()
		);

	/* Fit screen*/
	chat_FitScreen();
	chat_OnOrientationChange();
	chat_OnClickLogin();

	/* prevent chat message scrolling when nick list overlaps chat*/
	var domUsers = document.querySelector('.chat_users_container');
	var domUsersContent = domUsers.querySelector('.chat_users');
	domUsers.addEventListener('touchmove', function(e){
		if (domUsersContent.scrollHeight <=  domUsersContent.offsetHeight) e.preventDefault(); 
	}, false);
	
	/**-------------------------------------------------- LOGIN HANDLER ------------------------------------------------------------
	 * Performs request of login parameters for media server. 
	 * 
 	 * @param {String} sSalt
 	 * @param {Function} funcCallback
	 */
	function	loginHandler(sSalt, funcCallback)
	{
		$.get(ai_GetInfo('server_url') + 'mapp-listener/media_login/?' + 'login=' + sUserID + '&pw=' + ai_Get('pw') + '&salt=' + sSalt + '&id=' + sChannelID, funcCallback);
	}
};

/**
 * Extension of chat GUI for mobile application.
 */
function MobileChatExtension()
{	
	var domUsersListCont = $('#chat_wrapper .chat_users_container');
	var domUsers = $('#chat_users_container');
	$('#chat_wrapper .chat_history_container').click(function(){
			hideUsersList()
		});
	$('#chat_users_toggle').click(function(){
			if (domUsers.hasClass('expanded'))
			{
				hideUsersList();
			}
			else 
			{
				/* show users list*/
				domUsers.removeClass('hidden');
				domUsers.addClass('expanded');
				chatGUI.updateVisibility();
			}
		});

	$('.chat_users').click(function(e){ 
		if (!$(e.target).is('.chat_ban_btn, .chat_mute_btn'))
			hideUsersList(); 
	});

	/**
	 *	Hides panel of users list.
	 */
	function hideUsersList()
	{
		domUsers.removeClass('expanded');
		setTimeout(function(){ if (!domUsers.hasClass('expanded')) domUsers.addClass('hidden'); }, 350);
	};
	
	/**
	 *	Handler of chat login.
	 */
	this.onLogin = function()
	{
		$('.chat_bottom_intfc').css('display', 'block');
		$('.chat_bottom_login').css('display', 'none');
		$('#chat_users_toggle').removeClass('hidden');
	};

	/**
	 *	Handler of chat logout.
	 */
	this.onLogout = function()
	{
		$('.chat_bottom_intfc').css('display', 'none');
		$('.chat_bottom_login').css('display', 'block');
		$('#chat_users_toggle').addClass('hidden');
	};

	/**
	 *	Handler of private messsage notification.
	 */
	this.onPrivateMsgCount = function(user, count, total)
	{
		var domUser = domUsersListCont.find('.chat_user[data-id="' + user.id + '"]');

		/* Display number of private messages in the users list and in the private tabs*/
		if (domUser.length)
			domUser.find('.chat_pm_btn').text(count > 0 ? Math.min(count, 99) : '').toggleClass('active', count > 0);

		/* Change color of the users list toggle to show that there are some private messages*/
		$('#chat_users_toggle').toggleClass('private', total > 0);

		return true;
	};

	/**
	 *	On create private page.
	 */ 
	this.onCreatePrivatePage = function(user, domPageContainer, domTab)
	{
		chatGUI.closeAllPrivateTabs();
	};

	this.onPoll = function(bPoll)
	{
		$('.chat_smileys').toggleClass('shift_corner', bPoll);
	};

	/* Messages menu*/

	var msgMenu = null;

	function showCurMsgMenu()
	{
		var msgID = msgMenu.id;
		var msgIndex = msgMenu.index;
		window.plugins.listpicker.showPicker({
			title: msgMenu.text['menu_title'],
			items: msgMenu.items,
			doneButtonLabel: msgMenu.text['done_btn'],
			cancelButtonLabel: msgMenu.text['cancel_btn']
		},
			/* report selection handler*/
			function (itemValue) { 
				if (itemValue == 'report')
				{
					if (device.platform == "iOS")
						setTimeout(function () { showReportMenu(msgID, msgIndex); }, 1500);
					else
						showReportMenu(msgID, msgIndex);
				}
				else
					chatGUI.applyMessageAction(itemValue); 
			},
			function () { chatGUI.deselectMessage(); }
		);
		msgMenu = null;
	}

	this.showMsgMenu = function(msgID, msgInd, bDelete, bMyMsg, bBan)
	{
		var text = chat_LangMenu();
		var items = [];
		items.push({text: text[bDelete ? 'delete_msg' : 'hide_msg'], value: (bDelete ? 'delete' : 'hide')});
		if (!bMyMsg) items.push({text: text['report_msg'], value: 'report'});
		msgMenu = {id: msgID, index: (msgInd ? msgInd : 0), items: items, text: text};
		if (bBan) return;
		showCurMsgMenu();
		
	};

	this.onCheckBan = function(id, index, banned)
	{
		if (!msgMenu || (msgMenu.id != id && msgMenu.index != index)) return;
		msgMenu.items.push({text: msgMenu.text[banned ? 'unban_by_msg' : 'ban_by_msg'], value: (banned ? 'unban' : 'ban')});
		showCurMsgMenu();
	};

	/**
	 *	Shows ban dialog.
	 */
	this.showBanDlg = function(userID, nick, banned)
	{
		var text = chat_LangBanMenu();
		window.plugins.listpicker.showPicker({
			title: text['ban_title'].replace('%', nick),
			items: banned ?
				[
					{text: text['unban'], value: 'unban'}
				]
				:
				[
					 {text: text['ban_hour'], value: 'hour'}
					,{text: text['ban_day'], value: 'day'}
					,{text: text['ban_week'], value: 'week'}
				],
			doneButtonLabel: text['done_btn'],
			cancelButtonLabel: text['cancel_btn']
		},
			/* report selection handler*/
			function (itemValue) { 
				chatGUI.setUserBan(userID, itemValue); 
			},
			function () { /* Canceled */ }
		);
	};

	/**
	 *	Shows mute dialog.
	 */
	this.showMuteDlg = function(userID, nick, muted)
	{
		var text = chat_LangMuteMenu();
		var val = muted ? 'unmute' : 'mute';
		window.plugins.listpicker.showPicker({
			title: text['mute_title'].replace('%', nick),
			items: [ {text: text[val], value: val} ],
			doneButtonLabel: text['done_btn'],
			cancelButtonLabel: text['cancel_btn']
		},
			/* report selection handler*/
			function (itemValue) { 
				chatGUI.setUserMute(userID, itemValue == 'mute'); 
			},
			function () { /* Canceled */ }
		);
	};

	/**
	 * Shows report menu for a message.
	 *
	 * @param {any} msgID
	 * @param {any} msgIndex
	 */
	function showReportMenu(msgID, msgIndex)
	{
		var textItems = news_GetActionLang();
		window.plugins.listpicker.showPicker({
			title: textItems['report_title'],
			items: [
				{ text: textItems['insult'], value: 'insult' },
				{ text: textItems['adult_content'], value: 'adult_content' },
				{ text: textItems['drug_advocacy'], value: 'drug_advocacy' },
				{ text: textItems['child_porn'], value: 'child_porn' },
				{ text: textItems['extremism'], value: 'extremism' },
				{ text: textItems['suicide'], value: 'suicide' }
			],
			doneButtonLabel: textItems['done_btn'],
			cancelButtonLabel: textItems['cancel_btn']
		},
			/* report selection handler*/
			function (itemValue) {
				chatGUI.sendMsgReport(msgID, msgIndex, itemValue);
			},
			function () { chatGUI.deselectMessage(); }
		);
	}

	/**
	 *	Handler of report send.
	 */
	this.onMsgReportSent = function(msgID, msgIndex, res)
	{
		var text = chat_LangReportResult();
		navigator.notification.alert(text[res ? "report_success" : "report_failed"], function(){
			chatGUI.deselectMessage();
			},"");
	}
}

/*----------------------------- LANGUAGE ----------------------------------------------------------------*/
function chat_Lang()
{
	return {
		 tab_common: translate('Общий')
		,your_nick: translate('Ваше имя')
		,you_banned: translate('Вы забанены')
		,auth_required: translate('<u>Авторизуйтесь</u> чтобы написать в чат')

		,nick_already_exists: translate('Такой ник уже используется')
    	,cant_set_empty_nick: translate('Нельзя установить пустой ник')
    	,wrong_nick: translate('Недопустимое имя')
    	,msg_flood_detected: translate('Вы слишком часто отсылаете сообщения')
    	,target_user_not_found: translate('Пользователь не найден')
		,poll_question: translate('Вопрос')
		,poll_answers: translate('Варианты ответов')
		,poll_result: translate('Показывать результаты')
	  	,poll_start: translate('Начать опрос')
	  	,poll_stop: translate('Закончить опрос')
	  	,poll_new: translate('Новый опрос')
	};
}

function chat_LangBanMenu()
{
	return Object.assign(
	{
		 ban_title: translate('Забанить пользователя: %')
		,ban_hour: translate('На час')
		,ban_day: translate('На день')
		,ban_week: translate('На неделю')
		,unban: translate('Разбанить')
	}, chat_LangStdDlg());
}

function chat_LangMuteMenu()
{
	return Object.assign(
	{
		 mute_title: translate('Блокировать сообщения от пользователя: %')
		,mute: translate('Блокировать')
		,unmute: translate('Разблокировать')
	}, chat_LangStdDlg());
}

function chat_LangMenu()
{
	return Object.assign(
	{
		 menu_title: translate('Выберите действие')
		,done_btn: translate('Готово')
		,cancel_btn: translate('Отмена')
		,delete_msg: translate('Удалить сообщение')
		,hide_msg: translate('Скрыть сообщение')
		,report_msg: translate('Пожаловаться на сообщение')
		,ban_by_msg: translate('Забанить')
		,unban_by_msg: translate('Разбанить')
	}, chat_LangStdDlg());
}

function chat_LangStdDlg()
{
	return {
		 done_btn: translate('Готово')
		,cancel_btn: translate('Отмена')
	};
}

function chat_LangReportResult()
{
	return {
		 report_success: translate('Отчет отправлен')
		,report_failed: translate('Ошибка отправки отчета')
		,done_btn: translate('Готово')
	};
}

/*------------------------------------------------- CHAT HTML ----------------------------------------------------------*/
function chat_HTML( nChannelID )
{
	return `<div id="chat_channel_id" style="display:none">${nChannelID}</div>
<div id="chat_user_id" style="display:none">${ ai_Get('login') }</div>`;
}

/*------------------------------------------------------ ON CLICK LOGIN -----------------------------------------------------------*/
function chat_OnClickLogin()
{
	$('body').undelegate('.chat_bottom_login').delegate('.chat_bottom_login', 'click', function(){
		ai_SetPageURL( '', 'login', 'goback' );
		menu_Main();
		login_Main();
	});
}

/*-------------------------------------------------------- FIT SCREEEN ------------------------------------------------------------*/
function chat_FitScreen()
{
	if (!$("#vzchannel_info").length || !$("header").length || !$("#chat_wrapper").length) return;

	var nHeaderHeight = $("header").height();
	var nBottomMenu = $('#main_menu').height();
	var a = parseInt($("#vzchannel_info").css("margin-top"));
	var nChannelInfoHeight = parseInt($("#vzchannel_info").height() + parseInt($("#vzchannel_info").css("margin-top")) + parseInt($("#vzchannel_info").css("margin-bottom")));
	var nChatHeight = Math.floor(window.innerHeight - nHeaderHeight - nChannelInfoHeight - nBottomMenu - 1);

	$("#chat_wrapper").css("height", nChatHeight + "px");
}

/*---------------------------------------------------- ON  ORIENTATION CHANGE -----------------------------------------------------*/
function chat_OnOrientationChange()
{
	window.removeEventListener("resize", chat_FitScreen);
	window.addEventListener("resize", chat_FitScreen);
}